from . import BaseClientParser
from ua_extract.enums import AppType


class PIM(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.PIM

    fixture_files = [
        'upstream/client/pim.yml',
    ]


__all__ = [
    'PIM',
]



from .base import BaseDeviceParser
from ua_extract.enums import DeviceType
from ...lazy_regex import RegexLazy
from ...settings import BOUNDED_REGEX

FACEBOOK_NOTEBOOK_FRAGMENT = RegexLazy(BOUNDED_REGEX.format('FBMD/'))


class Notebook(BaseDeviceParser):
    __slots__ = ()
    DEVICE_TYPE = DeviceType.Desktop

    fixture_files = [
        'upstream/device/notebooks.yml',
    ]

    def _parse(self) -> None:
        if FACEBOOK_NOTEBOOK_FRAGMENT.search(self.user_agent):
            return super()._parse()


__all__ = [
    'Notebook',
]



from unittest import TestCase

from ua_extract.utils import (
    mostly_repeating_characters,
    only_numerals_and_punctuation,
    mostly_numerals,
    random_alphanumeric_string,
    uuid_like_name,
    normalize_app_name,
)


# -----------------------------------------------------------------------
class TestTrimSuffixes(TestCase):

    def test_trim_suffixes(self):
        for full_name, trimmed_name in (
            ('AirbnbNotificationServiceAppExtension', 'Airbnb'),
            ('NetworkingExtension', 'Networking'),
            ('OneDriveMessageExtension', 'OneDrive'),
            ('OneDriveShareExtension', 'OneDrive'),
            ('OneSignalNotificationServiceExtension', 'OneSignal'),
            ('RampNotificationServiceExtension', 'Ramp'),
            ('RampWidgetExtension', 'Ramp'),
            ('SafariNotificationAgent', 'Safari'),
            ('SafewayServiceExtension', 'Safeway'),
            ('SouthwestPushPreview', 'Southwest'),
            ('SouthwestWidgetExtension', 'Southwest'),
            ('SouthwestWidgetExtension', 'Southwest'),
            ('StocksKitService', 'Stocks'),
            ('StocksWidget', 'Stocks'),
            ('TWorldNotificationService', 'TWorld'),
            ('TeamSpace.Share-Extension', 'TeamSpace'),
            ('TeamSpaceShareExtension', 'TeamSpace'),
            ('VerseWidgetExtension', 'Verse'),
            ('WalmartWidgets', 'Walmart'),
            ('WyzeNotiService', 'Wyze'),
            ('com.agilebits.onepassword-ios.extension', 'com.agilebits.onepassword-ios'),
            ('com.apple.NewDeviceOutreach.Extension', 'com.apple.NewDeviceOutreach'),
            ('com.apple.SystemProfiler.AboutExtension', 'com.apple.SystemProfiler'),
            ('com.apple.SystemProfiler.IntentExtension', 'com.apple.SystemProfiler'),
            ('com.apple.SystemProfiler.VoiceFramework', 'com.apple.SystemProfiler'),
            ('com.apple.mobilenotes.SharingExtension', 'com.apple.mobilenotes'),
            ('com.costco.costco.notification-extension', 'com.costco.costco'),
            ('com.google.Drive.ExtensionFramework', 'com.google.Drive'),
            ('com.google.Drive.FileProviderExtension', 'com.google.Drive'),
            ('com.google.Gmail.AppAndExtensionsFramework', 'com.google.Gmail'),
            ('com.google.GoogleMobile.TrendsTodayExtension', 'com.google.GoogleMobile'),
            ('com.google.Maps.HomeTrafficWidgetExtension', 'com.google.Maps'),
            ('com.google.calendar.TodayExtension', 'com.google.calendar'),
            ('com.google.photos.ModuleFramework', 'com.google.photos'),
            ('onedrivefileprovider', 'onedrive'),
        ):
            self.assertEqual(normalize_app_name(full_name), trimmed_name)


# -----------------------------------------------------------------------
class TestNotUUIDname(TestCase):

    def test_user_agent_like_uuid(self):
        for ua in (
                '738FAAEF-30CF',
                '6BFAD903-A5EA-4E34',
                '5FAEB6ED-AE46-4A26-BA1B',
                '4U9WD-2TNWA-OQTSJ-QQDIG',
                '4UWWU-WBDXC-VYFN3-QDJMH',
                '4U4AP-Z24UV-BTW0P-QW93F',
                '4UY9X-NSVVL-O4BZQ-QIMCL-QTDCH',
                '4UNEF-PB63F-X9ZAS-AO3TB-HI9JO',
                'ea1866cb-c89a-6d5d-89b8-afdcdb715237',
                'a_05D42541-648B-41DD-B11F-0CAF61F4CE19-660-0000004D54BA56F2',
                'a_03E848FE-2BD9-4400-B87B-172C35D4A309-3121-00000B5202325FA0',
        ):
            self.assertTrue(
                uuid_like_name(ua),
                msg='%s is similar to a UUID' % ua,
            )

    def test_user_agent_not_like_uuid(self):
        for ua in (
                'Waterman-WZVN',
                'Grainger-Cats',
                'Boxcarts-8932',
                'Google-Java-Http-Lib',
        ):
            self.assertFalse(
                uuid_like_name(ua),
                msg='%s is not similar to a UUID' % ua,
            )


# -----------------------------------------------------------------------
class TestNotNumerals(TestCase):

    def test_user_agent_mostly_numerals(self):
        for ua in (
                '2349090',
                'Tx23490923v',
                '4444444444444444444444444sdfasd',
        ):
            self.assertTrue(
                mostly_numerals(ua),
                msg='%s is mostly numerical' % ua,
            )

    def test_numerical_names_spaces_allowed(self):
        """
        If a name with mostly numerals contains a space, don't reject it.
        """
        for ua in (
                'Xray 2.35.90/23.19',
                'Tesla 9-5883/2.35.90/23.19',
        ):
            self.assertFalse(
                mostly_numerals(ua),
                msg='%s is not considered mostly numerical' % ua,
            )

    def test_not_mostly_numerals_punctuation(self):
        for ua in (
                '234.90/90',
                '2-3.40/5',
                '345/90.9 45.20.4.5',
        ):
            self.assertTrue(
                only_numerals_and_punctuation(ua),
                msg='%s is only numerals and punctuation' % ua,
            )


class TestNotGibberish(TestCase):

    def test_gibberish_ua_strings(self):
        for ua in (
                '8MIj7Yq0J1VxqYRS kd 0x0ovW0CTyvXAAAAAA',
                'Ya0eXACkoylZb9TS9qacxCw8XxYYHI XQAAAAA',
                'o8HG9zLURJv6DKX1MyqnXbnJoycf/ rWwAAAAA',
                'Ya0eXACkoylZb9TS9qacxCw8XxYzWWuXQAAAAA',
                'Ya0eXACkoylZb9TS9qacxCw8XxYXW0gYQAAAAA',
                '001471FmBjtgZvahkMJdcGJhhjXuuD99',
                '002261f23pRwrjIEtS1MrX0lZ4hx8N7P',
                '002353ueFfucaEDjKRbKwlpDpecxwYwC',
                'ziNICEarE9VlaPSkhDAyZrkZSpuEkIA',
                '002921Fw0cheewpv5V82uy0sQ5DLeXyC',
                '002404WQDQjvKP3nVYqAhONTZ8U3IJOq',
                '003703Z6BUGEG0SGHIGyCbPoLC4CjTMw',
                '003098zRAFF8LcClf2Db1g1WPHsFhBOh',
                'ziNICEarE9VlaPSkhDAyZrkZSpuEkIA',
                'yN6OR5feZZJLWQ7ACkTkSFVAzhhd7g',
                '005398AzsR1apMZDU3R4eKUn2Ce2YP2t',
                '019075NAmsDfAyLQwfYzrItmHUokQmg4',
                '988629DFoPiphgYXrAXFQ4PVvARYJhGo',
                '008847Zab1zwuMkakvROGNpDADGmsncM',
                '0176150ybYVbSawNcofPImqRStnzfQNd',
                '0187803xp7UeBtUytFoixmJAUgKIDDpL',
                '029212UcpIeYpvvoYWWEjuzn4UxMTOVN',
                '988629DFoPiphgYXrAXFQ4PVvARYJhGo',
                '029607qypCjEVzUvbkocrIrLXEuVIeen',
                '04511892N6RXcmAMsIHTTUxwhynHDpAe',
                '016734gFTOFOHOVwkEEJiY2EUUeSRnZu',
                '124856aRNUaoXlJaDUtZULmL44LNoIqa',
                '1116610FoLlb56WHsQCrXfDMdmzClzpd',
                '166694kNMwYcM2daaPAdkqunFhojumdS',
                'zWOjZVKqdtwb2vdv2hvWpHzUGNY',
                'holav1_10593F39DD2DAEBC',
                'D6A8DBACB0C1',
                'HKUM_XBw3S',
                'HKUM_X7ebO2fs',
                'vVNYZaiXO9Hd5zAi',
                'W6o8uLSXW8gagZqDA',
                'zmV3SkWiLPNcSXNmhvFbQjpyk',
                '3vFsPvhg',
                '3VWsvq4e',
                '3vjQidb8',
                '21jWHcQV',
                '2ffAGSu1',
                '33Ji2XaL',
        ):
            self.assertTrue(
                random_alphanumeric_string(ua.lower()),
                msg='%s is random gibberish' % ua,
            )

    def test_not_gibberish_ua_strings(self):
        for ua in (
                'rss junkie',
                'Add Music',
                'Adobe GC',
                'ACSS Pro',
                'Adobe Bridge 2022',
                'Agilissystems Linxup',
                'Norwoodsystems Worldvoicemail',
                'Mediamushroom Copymydata2',
                'Mpixels Gravitynoodle',
                'Mizmowireless Acctmgt',
                'Shotgungaming Clayhuntstart',
                'Riffsy Riffsykeyboard',
                'Thegrizzlylabs Geniussign',
                'Peoplefun Wordflowers',
                'Pilottravelcenters Mypilot',
                'Salehotaibiapps Blackholecut',
                'Shareitagain Mymoodsmiley',
                'Sunnystudio Photovault',
                'Ag World',
                'HP-Inkjet-WebUpdate-1.1-wm',
                'Karaokemusicapplab Sing-Sharp',
                'Midasplayer Apps Bubblewitchsaga3',
                'NDES client 10.0.18362.1766',
                'X-Plane Installer 4.13r1-final-APL10.16',
                'Meevii--IOS-_640_1136_iPodtouch_iPod9,1_14.6_2.80.0_',
                'WeatherWeatherFoundation115F',
                'VrboTripWidgetExtension',
                'Wordswithfriends3',
                'Farmville2Countryescape',
                'yourswagfontsandquotesfree',
                'ZooskAndroid',
                'ZombodroidMemegenerator',
                'ZombieOutbreakSimulator',
                'WatercolorPaintingUniqueArt',
                'YellowPagesMessageApp',
                'ZeroiOSWidgetsExtension',
                'YouTubeUnplugged',
                'Youtube2016Oct16KI',
                'YouTubeMusicWidgetKitExtension',
                'YouTubeMusicTodayExtensionSkylark',
                'YahooWeatherNotificationServiceExtension',
                'YahooWeatherWeatherWidgetExtension',
                'YahooWeatherLiveWidget',
                'YAppNotificationServiceExtension',
                'YahooNewsWidgetExtension',
                'Volcanoes&Earthquakes',
                'MoviePosterWallpaperMaker',
                'LedWiFiDalsRGBW',
                'PoliceChase2Trucker',
                'PaperLabelMaker',
                'WorldTranslatorLite',
                'SynchronyCarCare',
                'AdobeCloudShare',
                'Americantruck2017',
                'worldtimebuddy',
                'AudioBook002',
                'simulator2018',
                'Scan2CadV10',
                'Perfect365',
                'StocksProTodayWidget',
                'bettycrockerbigbookcupcakes',
                'flytransporterairplanepilot',
                'Tube360VR',
                '3026moat',
                'goknots3d',
                'SuperBird',
                'wordcafe',
                'Vivaldi',
                'UT4Aplus',
                'CoffeeF2P',
                'zombiegun3d',
                'sniper3d',
                'TimeClockFree',
                'Volo',
                'Syncro',
                'SynkedUp',
                'simgai3d',
                'sylvaniasmart',
                'Tuunes',
                'Traffie',
                'Toyota',
                'talk',
                'tabcar',
                'Tapcart',
                'Snaptube',
                'Qlock',
                'Quickshot',
                'Roblox',
                'Scrapy',
                'Schoology',
                'Scape',
                'Scanbot',
                'Setup',
                'rampage',
                'PikLab',
                'PicLab',
                'Polyglot',
                'Popsa',
                'pamadunknbeat',
                'woodoku',
                'Wordly',
                'WeHunt',
                'Weebly',
                'weirdlaws',
                'WeDrum',
                'RackIPH',
                'Owltail',
                'Paycor',
                'PCAPro',
                'Roadie',
                'Quora',
                'ToonPackiMessage',
                'Voicy',
                'Wakie',
                'Wikibot',
                'WikiArt',
                'YumpuNews',
                'VOClock',
                'Volkswagen',
                'Wakeboard',
                'wartroops1917',
                'pic2shop',
                'zescape',
                'zipCAT',
                '100Lessons',
                'BoxCar 2.30',
                'JQSmartBand',
                'OfficialDropboxObjCSDKv2',
                'Ezze',
                'OPPO',
                'Wiko',
                'Asus',
                'KaiOS',
                'Huawei',
                'LG',
                'video.like',
                'snomM700',
                'sendori-client-win32',
                'Stocks101',
                'TuneFox2',
                'project205',
                'oilhunt2',
                'pacman256',
                'rcplane3',
                'loggly-log4net-appender',
                'sendori-client-win32',
        ):
            self.assertFalse(
                random_alphanumeric_string(ua.lower()),
                msg=f'{ua} is not random gibberish',
            )


class TestNotRepeatingCharacters(TestCase):

    def test_mostly_repeating_characters(self):
        for ua in (
                'b888888888888888888888x888888',
                '88888888888888888888888888888',
                'PUUUUUUUUUUUUUUUUXXXXXXXXX555',
                'AKAKAKAKAKAKAKAKAKA',
        ):
            self.assertTrue(
                mostly_repeating_characters(ua),
                msg=f'{ua} is mostly repeating characters',
            )

    def test_not_mostly_repeating_characters(self):
        for ua in (
                '1island',
                'cmeaom Android Radarym',
                'AivAndroidPlayer',
                'Alaskaairlines Android',
        ):
            self.assertFalse(
                mostly_repeating_characters(ua),
                msg=f'{ua} is not mostly repeating characters',
            )



from . import BaseClientParser
from ..parser import ENDSWITH_DARWIN
from ua_extract.enums import AppType


class MobileApp(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.MobileApp

    fixture_files = [
        'local/client/mobile_apps.yml',
        'upstream/client/mobile_apps.yml',
    ]

    def check_all_regexes(self) -> bool | list[str]:
        # Don't check ahocorasick for user agents like:
        # R/3.6.0 (ubuntu-16.04) R (3.6.0 x86_64-pc-linux-gnu x86_64 linux-gnu)
        if self.user_agent_lower.startswith("r/"):
            return True
        if check_all := super().check_all_regexes():
            return check_all
        if ENDSWITH_DARWIN.search(self.user_agent_lower):
            return True
        return False


__all__ = [
    'MobileApp',
]



from urllib.parse import unquote
from ..base import ParserBaseTest
from ...parser import VendorFragment


class TestVendorFragment(ParserBaseTest):

    fixture_files = [
        'tests/parser/fixtures/upstream/vendorfragments.yml',
    ]

    def test_parsing(self):
        fixtures = self.load_fixtures()

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('useragent'))
            expect = fixture['vendor']
            parsed = VendorFragment(self.user_agent, None).parse()
            self.assertEqual(expect, parsed.ua_data['brand'])


__all__ = [
    'TestVendorFragment',
]



from ..base import ParserBaseTest
from ...parser import Bot


class TestBot(ParserBaseTest):

    def test_get_info_from_ua_bot(self):
        ua = 'Googlebot/2.1 (http://www.googlebot.com/bot.html)'
        bot = Bot(ua=ua, client_hints=None)
        expected = {
            'name': 'Googlebot',
            'category': 'Search bot',
            'url': 'https://developers.google.com/search/docs/crawling-indexing/overview-google-crawlers',
            'producer': {
                'name': 'Google Inc.',
                'url': 'https://www.google.com/',
            },
            'version': '',
        }
        bot = bot.parse()
        bot.ua_data.pop('type', None)
        self.assertDictEqual(expected, bot.ua_data)

    def test_parse_no_bot(self):
        ua = 'Mozilla/4.0 (compatible; MSIE 9.0; Windows NT 6.1; SV1; SE 2.x)'
        bot = Bot(ua=ua, client_hints=None)
        bot.parse()
        self.assertEqual(bot.ua_data, {})


__all__ = [
    'TestBot',
]



from .base import BaseDeviceParser
from ua_extract.enums import DeviceType

# Extracted from ^(?:chrome|firefox|Abcd|Dark|KvshClient|Node.js|Report Runner|url|Zeus|ZmEu)$
GENERIC_BOT_UAS = {
    'chrome',
    'firefox',
    'abcd',
    'dark',
    'kvshclient',
    'node.js',
    'report runner',
    'url',
    'zeus',
    'zmeu',
    'google',
}


class Bot(BaseDeviceParser):
    __slots__ = ()
    DEVICE_TYPE = DeviceType.Unknown

    fixture_files = [
        'upstream/bots.yml',
    ]

    def check_all_regexes(self) -> bool | list[str]:
        if check_all := super().check_all_regexes():
            return check_all
        return self.user_agent_lower in GENERIC_BOT_UAS

    def is_bot(self) -> bool:
        return self.matched_regex is not None

    def set_details(self) -> None:
        return super().set_details() if self.is_bot() else None


__all__ = [
    'Bot',
]



from . import BaseClientParser
from ua_extract.enums import AppType


class DesktopApp(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.DesktopApp

    fixture_files = [
        'local/client/desktop_apps.yml',
    ]


class OsUtility(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.OsUtility

    fixture_files = [
        'local/client/osutility.yml',
    ]


class Antivirus(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.Antivirus

    fixture_files = [
        'local/client/antivirus.yml',
    ]


__all__ = (
    'DesktopApp',
    'Antivirus',
    'OsUtility',
)



try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum  # type: ignore[no-redef]


class AppType(StrEnum):
    Antivirus = 'av'
    ArtificialIntelligence = 'ai'
    Browser = 'browser'
    FeedReader = 'feed reader'
    Game = 'game'
    Library = 'library'
    MediaPlayer = 'mediaplayer'
    Messaging = 'messaging'
    Navigation = 'navigation'
    OsUtility = 'osutility'
    Productivity = 'productivity'
    P2P = 'p2p'
    PIM = 'pim'
    VpnProxy = 'vpnproxy'

    # Generic types
    DesktopApp = 'desktop app'
    MobileApp = 'mobile app'
    Generic = 'generic'
    Unknown = ''


class DeviceType(StrEnum):
    Desktop = 'desktop'
    Smartphone = 'smartphone'
    Tablet = 'tablet'
    FeaturePhone = 'feature phone'
    Console = 'console'
    TV = 'tv'  # including set-top boxes, Blu-ray players
    CarBrowser = 'car browser'
    SmartDisplay = 'smart display'
    Camera = 'camera'
    PortableMediaPlayer = 'portable media player'
    Phablet = 'phablet'
    SmartSpeaker = 'smart speaker'
    Wearable = 'wearable'  # including set watches, headsets
    Peripheral = 'peripheral'  # including portable terminal, portable projector
    Unknown = ''



from .parser import Parser
from .os_fragment import OSFragment
from ..lazy_regex import RegexLazyIgnore
from .settings import (
    normalized_name,
    CHROME_OS_MODELS,
    DESKTOP_OS,
    OPERATING_SYSTEMS,
    OS_TO_ABBREV,
    OS_FAMILIES,
    FAMILY_FROM_OS,
    ANDROID_APPS,
    FIREOS_VERSION_MAPPING,
    LINEAGEOS_VERSION_MAPPING,
)
from ..settings import BOUNDED_REGEX

WEBOS_VERSION = RegexLazyIgnore(r'WEBOS([\d\.]+)')

ARM_REGEX = RegexLazyIgnore(
    BOUNDED_REGEX.format(r'arm[ _;)ev]|.*arm$|.*arm64|aarch64|Apple ?TV|Watch ?OS|Watch1,[12]')
)
LOONGARCH_REGEX = RegexLazyIgnore(BOUNDED_REGEX.format('loongarch64'))
MIPS_REGEX = RegexLazyIgnore(BOUNDED_REGEX.format('mips'))
SPARK_REGEX = RegexLazyIgnore(BOUNDED_REGEX.format('sparc64'))
SUPERH_REGEX = RegexLazyIgnore(BOUNDED_REGEX.format('sh4'))
WINDOWS_REGEX = RegexLazyIgnore(
    BOUNDED_REGEX.format('64-?bit|WOW64|(?:Intel)?x64|WINDOWS_64|win64|IRIX;?64|.*amd64|.*x86_?64')
)
x86_REGEX = RegexLazyIgnore(BOUNDED_REGEX.format('.*32bit|.*win32|(?:i[0-9]|x)86|i86pc'))


class OS(Parser):
    fixture_files = [
        'local/oss.yml',
        'upstream/oss.yml',
    ]

    DESKTOP_OS = DESKTOP_OS
    OPERATING_SYSTEMS = OPERATING_SYSTEMS
    OS_TO_ABBREV = OS_TO_ABBREV
    OS_FAMILIES = OS_FAMILIES
    FAMILY_FROM_OS = FAMILY_FROM_OS

    def is_desktop(self) -> bool:
        if self.client_hints and self.client_hints.mobile:
            return self.client_hints.is_desktop()
        return self.family() in self.DESKTOP_OS

    def short_name(self) -> str:
        return self.ua_data.get('short_name') or ''

    def family(self) -> str:
        return self.ua_data.get('family') or ''

    def is_known(self) -> bool:
        if self.short_name() == self.UNKNOWN:
            return False
        return super().is_known()

    def check_all_regexes(self) -> bool | list[str]:
        if check_all := super().check_all_regexes():
            return check_all
        return self.is_ios_fragment()

    def platform(self) -> str:
        if ch := self.client_hints:
            ch_architecture = self.client_hints.architecture.lower()
            for fragment, code in (
                ('arm', 'ARM'),
                ('loongarch64', 'LoongArch64'),
                ('mips', 'MIPS'),
                ('sh4', 'SuperH'),
                ('sparc64', 'SPARC64'),
            ):
                if fragment in ch_architecture:
                    return code

            if 'x64' in ch_architecture or (ch.bitness == '64' and 'x86' in ch_architecture):
                return 'x64'
            if ch_architecture == 'x86':
                return 'x86'
            return ch_architecture

        user_agent = self.user_agent
        for platform_regex, code in (
            (ARM_REGEX, 'ARM'),
            (LOONGARCH_REGEX, 'LoongArch64'),
            (MIPS_REGEX, 'MIPS'),
            (SUPERH_REGEX, 'SuperH'),
            (SPARK_REGEX, 'SPARC64'),
            (WINDOWS_REGEX, 'x64'),
            (x86_REGEX, 'x86'),
        ):
            if platform_regex.search(user_agent):
                return code

        return ''

    def _parse(self) -> None:
        super()._parse()
        if not self.ua_data:
            OSFragment(self.user_agent, self.client_hints).parse()

    def set_details(self) -> None:
        super().set_details()
        os_from_ua = self.parse_os_from_useragent()
        os_from_hints = self.parse_os_from_client_hints()

        name_from_ua = os_from_ua.get('name') or ''
        short_name_from_ua = os_from_ua.get('short_name') or ''
        family_from_ua = os_from_ua.get('family') or ''
        version_from_ua = os_from_ua.get('version') or ''

        if os_from_hints:
            name = os_from_hints.get('name') or ''
            short = os_from_hints.get('short_name') or short_name_from_ua
            family = os_from_hints.get('family') or family_from_ua
            version = os_from_hints.get('version') or ''
            ch_model = self.client_hints.model if self.client_hints else ''

            # Use version from user agent if non was provided in client hints,
            # but os family from useragent matches
            if not version and family_from_ua == family:
                version = version_from_ua

            if name == 'Windows' and version and version[0] == '0':
                version = '' if version_from_ua == '10' else version_from_ua

            # If the OS name detected from client hints matches the OS family from user agent
            # but the os name is another, use the one from user agent, as it might be more detailed
            if name == family_from_ua and name != name_from_ua:
                name = name_from_ua

                if name == 'LeafOS' or name == 'HarmonyOS':
                    version = ''
                elif name == 'PICO OS':
                    version = version_from_ua
                elif name == 'Fire OS' and os_from_hints:
                    major_version = version.split('.')[0] or '0'
                    version = FIREOS_VERSION_MAPPING.get(version) or FIREOS_VERSION_MAPPING.get(
                        major_version, version_from_ua
                    )

            # Chrome OS is in some cases reported as Linux in client hints,
            # we fix this only if the version matches
            if name == 'GNU/Linux' and name_from_ua == 'Chrome OS' and version == version_from_ua:
                name = name_from_ua
                short = short_name_from_ua

            # Chrome OS is in some cases reported as Android in client hints
            if name == 'Android' and (name_from_ua == 'Chrome OS' or ch_model in CHROME_OS_MODELS):
                name = 'Chrome OS'
                version = ''
                short = 'COS'

            # Meta Horizon is reported as Linux in client hints
            if name == 'GNU/Linux' and name_from_ua == 'Meta Horizon':
                name = name_from_ua
                short = short_name_from_ua
        elif name_from_ua:
            name = name_from_ua
            family = family_from_ua
            version = version_from_ua
            short = short_name_from_ua
        else:
            return

        if ch := self.client_hints:
            if ch.app in ANDROID_APPS and name != 'Android':
                name = 'Android'
                family = 'Android'
                short = 'ADR'
                version = ''
            elif ch.app == 'org.lineageos.jelly' and name != 'Lineage OS':
                major_version = version.split('.')[0] or '0'
                name = 'Lineage OS'
                family = 'Android'
                short = 'LEN'
                version = (
                    LINEAGEOS_VERSION_MAPPING.get(version)
                    or LINEAGEOS_VERSION_MAPPING.get(major_version)
                    or ''
                )
            elif ch.app == 'org.mozilla.tv.firefox' and name != 'Fire OS':
                major_version = version.split('.')[0] or '0'
                name = 'Fire OS'
                family = 'Android'
                short = 'FIR'
                version = (
                    FIREOS_VERSION_MAPPING.get(version)
                    or FIREOS_VERSION_MAPPING.get(major_version)
                    or ''
                )

        self.ua_data |= {
            'name': name,
            'short_name': short,
            'version': version,
            'platform': self.platform(),
            'family': family,
        }

    def parse_os_from_client_hints(self) -> dict[str, str]:
        """
        Returns the OS that can be safely detected from client hints
        """
        if not (ch := self.client_hints):
            return {}

        os_data = {}
        if ch.architecture:
            os_data['platform'] = self.platform()
        if ch.platform_version:
            os_data['version'] = ch.platform_version
        if ch.platform:
            name_from_ua = os_data.get('name')
            if not name_from_ua or name_from_ua == 'Linux':
                os_data |= {
                    'name': ch.platform,
                    'family': self.FAMILY_FROM_OS.get(ch.platform, ch.platform),
                    'short_name': self.OS_TO_ABBREV.get(ch.platform.lower(), ''),
                }

        if not os_data and ch.client_hints_map:
            for os, version in ch.client_hints_map.items():
                osl = os.lower()
                if abbreviation := self.OS_TO_ABBREV.get(osl):
                    os_data = {
                        'name': os,
                        'short_name': abbreviation,
                        'version': version,
                        'family': self.FAMILY_FROM_OS.get(abbreviation, ''),
                    }
                    break

        if os_data:
            os_name = self.ua_data.get('name')
            version = os_data.get('version', '')

            if os_name == 'Windows':
                os_version = version.split('.')
                major_version = os_version[0] or '0'
                minor_version = '0' if len(os_version) == 1 else os_version[1]

                if major_version == '0':
                    minor_version_map = {'1': '7', '2': '8', '3': '8.1'}
                    version = minor_version_map.get(minor_version, version)
                elif '0' < major_version < '11':
                    version = '10'
                elif major_version > '10':
                    version = '11'

            # On Windows, version 0.0.0 can be either 7, 8 or 8.1, so we return 0.0.0
            elif os_name != 'Windows' and version == '0':
                version = ''

            os_data['version'] = version
            self.ua_data |= os_data

        return os_data

    def parse_os_from_useragent(self) -> dict[str, str | None]:
        """
        Returns the OS that can be detected from useragent
        """
        if self.ua_data:
            name = normalized_name(
                self.ua_data['name'].lower(),
                self.OS_TO_ABBREV,
                self.OPERATING_SYSTEMS,
            )
            abbreviation = self.OS_TO_ABBREV.get(name.lower(), self.UNKNOWN)
            return {
                'name': name,
                'short_name': abbreviation,
                'family': self.FAMILY_FROM_OS.get(abbreviation),
                'version': self.set_version(self.ua_data.get('version', '')),
            }
        return {}

    def set_version(self, version: str) -> str:
        version = super().set_version(version)

        if not version:
            # Extract version from WEBOS4.5 substring
            name = self.ua_data.get('name', '')
            if name and name.lower() == 'webos':
                match = WEBOS_VERSION.search(self.user_agent)
                if match:
                    version = match.group(1)

        return version


__all__ = (
    'DESKTOP_OS',
    'OS',
    'OPERATING_SYSTEMS',
    'OS_FAMILIES',
)



from ..base import DetectorBaseTest


class TestDetectTablet(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet.yml',
    ]


class TestDetectTablet1(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-1.yml',
    ]


class TestDetectTablet2(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-2.yml',
    ]


class TestDetectTablet3(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-3.yml',
    ]


class TestDetectTablet4(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-4.yml',
    ]


class TestDetectTablet5(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-5.yml',
    ]


class TestDetectTablet6(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-6.yml',
    ]


class TestDetectTablet7(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-7.yml',
    ]


class TestDetectTablet8(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-8.yml',
    ]


class TestDetectTablet9(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-9.yml',
    ]


class TestDetectTablet10(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-10.yml',
    ]


class TestDetectTablet11(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-11.yml',
    ]


class TestDetectTablet12(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/tablet-12.yml',
    ]



from .base import BaseDeviceParser
from ua_extract.enums import DeviceType
from ..parser import ENDSWITH_DARWIN, IPHONE_ONLY_UA
from ...lazy_regex import RegexLazy, RegexLazyIgnore
from .vendor_fragment import VendorFragment
from ...settings import BOUNDED_REGEX
from ..settings import ALWAYS_DESKTOP_OS, DESKTOP_OS

CHROME_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Chrome/[.0-9]*'))
CHROME_MOBILE_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'(?:Mobile|eliboM)'))
DESKTOP_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'(?:Windows (?:NT|IoT)|X11; Linux x86_64)'))
EXCLUDED_DESKTOP_FRAGMENT = RegexLazy(
    BOUNDED_REGEX.format(
        'CE-HTML|'
        ' Mozilla/|Andr[o0]id|Tablet|Mobile|iPhone|Windows Phone|ricoh|OculusBrowser|'
        'PicoBrowser|Lenovo|compatible; MSIE|Trident/|Tesla/|XBOX|FBMD/|ARM; ?([^)]+)'
    )
)
ANDROID_MOBILE_FRAGMENT = RegexLazy(
    BOUNDED_REGEX.format(r'Android( [.0-9]+)?; Mobile;|.*\-mobile$')
)
ANDROID_VRF_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Android( [.0-9]+)?; Mobile VR;| VR '))
ANDROID_TABLET_FRAGMENT = RegexLazy(
    BOUNDED_REGEX.format(r'Android( [.0-9]+)?; Tablet;|Tablet(?! PC)|.*\-tablet$')
)
ANDROID_TV_FRAGMENT = RegexLazyIgnore(
    r'Andr0id|(?:Android(?: UHD)?|Google) TV|\(lite\) TV|BRAVIA|Firebolt| TV$'
)
TOUCH_FRAGMENT = RegexLazy(BOUNDED_REGEX.format('Touch'))
GENERAL_TABLET_FRAGMENT = RegexLazy(BOUNDED_REGEX.format('HTC|Kindle'))
TV_FRAGMENT = RegexLazyIgnore(BOUNDED_REGEX.format(r'TV Store|WhaleTV|Smart TVS'))
TV_MINI_FRAGMENT = RegexLazyIgnore(BOUNDED_REGEX.format(r'\(TV;'))
TIZEN_TV_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'SmartTV|Tizen.+ TV .+$'))

# Some UA contain the fragment 'Pad/APad', so we assume those devices as tablets
PAD_TABLET_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Pad/APad'))

PUFFIN_DESKTOP_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Puffin/(?:\d+[.\d]+)[LMW]D'))
PUFFIN_PHONE_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Puffin/(?:\d+[.\d]+)[AIFLW]P'))
PUFFIN_TABLET_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'Puffin/(?:\d+[.\d]+)[AILW]T'))

OPERA_TABLET_FRAGMENT = RegexLazy(BOUNDED_REGEX.format('Opera Tablet'))
OPERA_TV_FRAGMENT = RegexLazy(BOUNDED_REGEX.format('Opera TV Store| OMI/'))

ENDSWITH_FIREFOX = RegexLazyIgnore(r'(Firefox|Iceweasel|Phoenix)/(?:\d+[.\d]+)$')
UA_CLIENT_HINTS_FRAGMENT = RegexLazyIgnore(
    r'Android (?:1[0-6][.\d]*; K(?: Build/|[;)])|1[0-6]\)) AppleWebKit'
)
TELEGRAM_ANDROID = RegexLazy('Telegram-Android/')


class Device(BaseDeviceParser):
    """
    This class should be the final device-type class checked.

    It should not be subclassed!
    """

    __slots__ = ()

    DEVICE_TYPE = DeviceType.Smartphone

    fixture_files = [
        'upstream/device/mobiles.yml',
    ]

    def check_all_regexes(self) -> bool | list[str]:
        # Match relatively generic UAs like:
        # UCWEB/2.0 (MIDP-2.0; U; zh-CN; IQ4406) U2/1.0.0 UCBrowser/3.4.3.532 U2/1.0.0 Mobile
        # UCWEB/2.0 (Linux; U; Opera Mini/7.1.32052/30.3697; en-US; LG-E405) U2/1.0.0 UCBrowser/8.8.1.359 Mobile
        if self.user_agent_lower.endswith(' mobile'):
            return True

        if self.client_hints:
            if self.client_hints.model:
                return True
            if self.has_user_agent_client_hints_fragment():
                return False

        if ac_match := super().check_all_regexes():
            return ac_match

        if self.user_agent_lower.startswith('iphone'):
            return IPHONE_ONLY_UA.match(self.user_agent) is not None

        if ENDSWITH_FIREFOX.search(self.user_agent) is not None:
            return True

        if ENDSWITH_DARWIN.search(self.user_agent) is not None:
            return True

        return self.user_agent_lower == 'msdw'

    def has_user_agent_client_hints_fragment(self) -> bool:
        if UA_CLIENT_HINTS_FRAGMENT.search(self.user_agent):
            if not TELEGRAM_ANDROID.search(self.user_agent):
                return True
        return False

    def _parse(self) -> None:
        """
        Loop through all brands of all device types trying to find
        a model. Returns the first device with model info.
        """
        if not (ac_matched := self.check_all_regexes()):  # noqa
            return

        ch = self.client_hints
        ch_model = ch.model if ch else None

        # ------------------------------------------------
        # Complete copy of the superclass _parse method
        for ua_data in self.regex_list:
            if self.known:
                break
            if matched := ua_data['regex'].search(self.user_agent):
                self.matched_regex = matched
                self.ua_data |= {k: v for k, v in ua_data.items() if k != 'regex'}
                self.known = True
            elif ch_model:
                main_fixture_dtype = ua_data.get('device')
                if ua_models := ua_data.get('models', []):
                    for model_data in ua_models:
                        if self.known:
                            break
                        model_fixture_dtype = model_data.get('device', main_fixture_dtype)
                        if not compatible_device_type(model_fixture_dtype, self.DEVICE_TYPE):
                            continue
                        matched = model_data['regex'].search(ch_model)
                        if matched:
                            self.matched_regex = matched
                            self.known = True
                            ua_data = {
                                k: v for k, v in ua_data.items() if k != 'regex' and k != 'models'
                            }
                            ua_data['model'] = model_data['model']
                            ua_data['device'] = model_fixture_dtype
                            self.ua_data = ua_data

                elif main_fixture_dtype == self.DEVICE_TYPE and (
                    matched := ua_data['regex'].search(ch_model)
                ):
                    self.matched_regex = matched
                    self.ua_data |= {k: v for k, v in ua_data.items() if k != 'regex'}
                    self.known = True
        # ------------------------------------------------

        if not self.ua_data and ch:
            self.ua_data |= {
                'type': ch.device_type(),
                'model': ch_model,
                'brand': 'Apple' if ch.platform == 'Mac' else '',
            }

        if not self.ua_data.get('brand'):
            # If no brand info was found, check known fragments
            vendor_fragment = VendorFragment(self.user_agent, self.client_hints).parse().ua_data
            if vendor_fragment:
                self.ua_data |= vendor_fragment

        if device_type := self.dtype():
            self.ua_data['type'] = device_type

        # Uncomment lines for debugging.
        # if ac_matched and not isinstance(ac_matched, bool):
        #     print(f'{self.cache_name}: Unwanted AC Match is {ac_matched}')

    def is_tablet(self) -> bool:
        """
        Check for various tablet fragments.
        """
        return PAD_TABLET_FRAGMENT.search(self.user_agent) is not None

    def check_android_device(
        self,
        dtype: DeviceType | str,
        os_name: str,
        os_version: str,
    ) -> DeviceType | None:
        """
        Chrome on Android passes the device type based on the keyword 'Mobile'
        If it is present the device should be a smartphone, otherwise it's a tablet
        See https://developer.chrome.com/multidevice/user-agent#chrome_for_android_user_agent
        """
        if os_name not in ('Android', 'MocorDroid'):
            return None

        # All detected feature phones running Android are more likely a smartphone
        if dtype == DeviceType.FeaturePhone:
            return DeviceType.Smartphone

        if not self.matched_regex:
            # All devices containing VR fragment are assumed to be a wearable
            if ANDROID_VRF_FRAGMENT.search(self.user_agent) is not None:
                return DeviceType.Wearable

            # Some UA contain the fragment 'Android; Tablet;' or 'Opera Tablet',
            # so we assume those devices as tablets
            if (
                ANDROID_TABLET_FRAGMENT.search(self.user_agent) is not None
                or OPERA_TABLET_FRAGMENT.search(self.user_agent) is not None
            ):
                return DeviceType.Tablet

            if ANDROID_MOBILE_FRAGMENT.search(self.user_agent) is not None:
                return DeviceType.Smartphone

            # Chrome on Android passes the device type based on the keyword 'Mobile'
            # If it is present the device should be a smartphone, otherwise it's a tablet
            # See https://developer.chrome.com/multidevice/user-agent#chrome_for_android_user_agent
            # Note: We do not check for browser (family) here, as there might be mobile apps using
            #       Chrome, that won't have a detected browser, but can still be detected.
            #       So we check the useragent for Chrome instead.
            if CHROME_FRAGMENT.search(self.user_agent) is not None:
                if CHROME_MOBILE_FRAGMENT.search(self.user_agent) is not None:
                    return DeviceType.Smartphone
                return DeviceType.Tablet

            # Android up to 3.0 was designed for smartphones only. But as 3.0, which was tablet
            # only, was published too late, there were a bunch of tablets running with 2.x
            #
            # With 4.0 the two trees were merged and it is for smartphones and tablets
            # So were are expecting that all devices running Android < 2 are smartphones
            # Devices running Android 3.X are tablets. Device type of Android 2.X and 4.X+
            # are unknown
            if not os_version:
                return None

            try:
                if float(os_version) < 2.0:
                    return DeviceType.Smartphone
                if 3.0 <= float(os_version) < 4.0:
                    return DeviceType.Tablet
            except (ValueError, TypeError):
                pass

        return None

    def check_puffin_device(self) -> DeviceType | None:
        if self.matched_regex:
            return None

        # All devices running Puffin Secure Browser that contain
        # letter 'D' are assumed to be desktops
        if PUFFIN_DESKTOP_FRAGMENT.search(self.user_agent) is not None:
            return DeviceType.Desktop

        # All devices running Puffin Web Browser that contain
        # letter 'P' are assumed to be smartphones
        if PUFFIN_PHONE_FRAGMENT.search(self.user_agent) is not None:
            return DeviceType.Smartphone

        # All devices running Puffin Web Browser that contain
        # letter 'T' are assumed to be tablets
        if PUFFIN_TABLET_FRAGMENT.search(self.user_agent) is not None:
            return DeviceType.Tablet

        return None

    def is_television(self, os_name: str) -> bool:
        """
        Check all the ways a device might be a television.
        """
        # All devices running Coolita OS are assumed to be a tv
        if os_name == 'Coolita OS':
            return True

        # All devices running Opera TV Store are assumed to be a tv
        if OPERA_TV_FRAGMENT.search(self.user_agent) is not None:
            return True

        if ANDROID_TV_FRAGMENT.search(self.user_agent) is not None:
            return True

        # All devices running Tizen TV or SmartTV are assumed to be a tv
        if TIZEN_TV_FRAGMENT.search(self.user_agent) is not None:
            return True

        # All devices containing TV fragment are assumed to be a tv
        if TV_MINI_FRAGMENT.search(self.user_agent) is not None:
            return True

        if TV_FRAGMENT.search(self.user_agent) is not None:
            return True

        return False

    def is_desktop(self, os_name: str) -> bool:
        if os_name in ALWAYS_DESKTOP_OS:
            return True

        if os_name not in DESKTOP_OS or GENERAL_TABLET_FRAGMENT.search(self.user_agent) is not None:
            return False

        # Returns if the parsed UA contains the 'Windows NT;' or 'X11; Linux x86_64' fragments
        fragment = (
            DESKTOP_FRAGMENT.search(self.user_agent) is not None
            and EXCLUDED_DESKTOP_FRAGMENT.search(self.user_agent) is None
        )
        return fragment or os_name == 'Mac' or ' Desktop' in self.user_agent

    def device_runs_feature_phone_os(self, os_name: str) -> bool:
        if not self.matched_regex and os_name == 'Java ME':
            return True

        # All devices running KaiOS are more likely feature phones
        return os_name == 'KaiOS'

    def is_windows_tablet(self, os_name: str, os_version: str) -> bool:
        """
        According to https://msdn.microsoft.com/en-us/library/ie/hh920767(v=vs.85).aspx
        Internet Explorer 10 introduces the "Touch" UA string token. If this token is present
        at the end of the UA string, the computer has touch capability, and is running Windows 8
        (or later). This UA string will be transmitted on a touch-enabled system running
        Windows 8 (RT)

        As most touch enabled devices are tablets and only a smaller part are desktops/notebooks
        we assume that all Windows 8 touch devices are tablets.
        """
        if self.matched_regex:
            return False
        if os_name == 'Windows RT' or (os_name == 'Windows' and os_version.startswith('8')):
            return TOUCH_FRAGMENT.search(self.user_agent) is not None
        return False

    def dtype(self) -> DeviceType | str:
        """Calculate Device Type, based on various parameters"""
        default_dtype = super().dtype()
        fixture_dtype = self.device_type_from_fixture()

        os_name = self.os_details.get('name', '')
        os_family = self.os_details.get('family', '')
        os_version = self.os_details.get('version', '')

        if self.device_runs_feature_phone_os(os_name):
            return DeviceType.FeaturePhone

        if self.client_hints and self.client_hints.is_television():
            return DeviceType.TV

        if android_device := self.check_android_device(fixture_dtype, os_name, os_version):
            return android_device

        if self.is_windows_tablet(os_name, os_version):
            return DeviceType.Tablet

        if device_from_puffin := self.check_puffin_device():
            return device_from_puffin

        if default_dtype != DeviceType.TV and self.is_television(os_name):
            return DeviceType.TV

        if default_dtype != DeviceType.Tablet and self.is_tablet():
            return DeviceType.Tablet

        if default_dtype != DeviceType.Desktop and self.is_desktop(os_name):
            return DeviceType.Desktop

        # Set device type to desktop for all devices running a
        # desktop OS that were not detected as another device type
        if not self.matched_regex and os_family in DESKTOP_OS:
            return DeviceType.Desktop

        if not self.user_agent:
            return DeviceType.Unknown

        return fixture_dtype or default_dtype

    def model(self) -> str:
        return self.ua_data.get('model', '')

    def __str__(self) -> str:
        return f'{self.model()} {self.dtype()}'


def compatible_device_type(actual: str, target: str) -> bool:
    """
    When iterating over model regexes, we need to skip regexes
    where the device type isn't compatible with the class being checked.
    """
    if actual == target:
        return True

    if target == DeviceType.Smartphone:
        return actual == DeviceType.Tablet or actual == DeviceType.Phablet

    return False


__all__ = [
    'Device',
]



"""
device_detector Test Cases. Run entire suite by:

python3 -m unittest

Run individual test class by:

python3 -m unittest ua_extract.tests.parser.test_bot
"""
try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum
from urllib.parse import unquote
import unittest
import yaml
try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from ua_extract.parser import ClientHints
from ..settings import ROOT
from .. import DeviceDetector


# App names -> Application ID map so that upstream
# test fixtures can pass without modifications
APPID_TO_NAME = {
    'com.google.GooglePlus': 'Google Plus',
    'com.google.android.youtube': 'YouTube',
    'com.google.android.apps.magazines': 'Google Play Newsstand',
    'WAP Browser': '',
}


class TestInvalidUserAgents(unittest.TestCase):

    def test_all_punctuation(self):
        """
        Bogus strings should not crash parser, and should return no client details
        """
        for ua in (
                '%7C%7C%27',
                ']]>><',
        ):
            dd = DeviceDetector(ua).parse()
            self.assertNotIn('client', dd.all_details)


class Base(unittest.TestCase):

    fixture_files = []
    Parser = None

    def assertEqual(self, first, second, msg=None, **kwargs):
        if first == 'None':
            first = None
        if second == 'None':
            second = None
        elif isinstance(second, StrEnum):
            second = second.value

        if isinstance(first, str) and isinstance(second, int):
            second = str(second)

        # Count all falsy types as equal
        if not first and (not second or second == '0'):
            return

        if not msg and hasattr(self, 'user_agent'):
            field = kwargs.get('field')
            ua_msg = f'\n\nFailed to parse {getattr(self, "user_agent")!r}\n'

            if field:
                field_msg = f'Field {field!r} expected value {first!r} != Parsed value {second!r}.'
            else:
                field_msg = f'Expected value {first!r} != Parsed value {second!r}'

            msg = f'{ua_msg} {field_msg}'

        super().assertEqual(first, second, msg)

    def load_fixtures(self):
        fixtures = []
        for ffile in self.fixture_files:
            with open(f'{ROOT}/{ffile}', 'r') as r:
                fixtures.extend(yaml.load(r, SafeLoader))
        return fixtures


class DetectorBaseTest(Base):
    Parser = DeviceDetector

    def confirm_client_name(self, fixture, parsed_name):
        """
        Client name may be specified as empty string in Matomo
        regexes, but we want to extract a generic name from the
        beginning of the regex.

        :parsed_name: The name the parser extracted
        """
        fixture_name = self.get_value(fixture, 'client', 'name')

        # if there's a fixture name, then the parsed value should match
        if fixture_name:
            self.assertEqual(fixture_name, parsed_name, field='client_name')

    def confirm_client_type(self, fixture, parsed_value):
        """
        Matomo will call many more apps "browsers" than we do.

        :parsed_value: The value the parser extracted
        """
        fixture_value = self.get_value(fixture, 'client', 'type')

        # like our name better...
        if fixture_value == 'browser' and fixture_value != parsed_value:
            return

        if fixture_value:
            self.assertEqual(str(fixture_value), str(parsed_value), field='client_type')

    def confirm_version(self, fixture, parsed_value):
        """
        If fixture data is empty, make sure the version string
        is a substring of the user agent.

        :parsed_value: The value the parser extracted
        """
        fixture_value = self.get_value(fixture, 'client', 'version')

        # if the fixture value specifies a version, it should match
        # otherwise, skip checking. The generic version extractor
        # will push/pull values a bit.
        if fixture_value:
            self.assertEqual(str(fixture_value), str(parsed_value), field='client_version')

    def matches_fixture_or_generic(self, fixture, key1, key2, parsed_value):
        """
        If fixture data is empty, allow the parsed_value to be "generic"
        which might be what our generic extractors return.

        :parsed_name: The name the parser extracted
        """
        fixture_value = self.get_value(fixture, key1, key2)

        if fixture_value:
            self.assertEqual(fixture_value, parsed_value)
        elif parsed_value:
            # check various generic values
            self.assertIn(parsed_value, ('generic', 'Windows', 'smartphone', 'XX', 'Unknown'))

    def get_value(self, fixture, key1, key2):
        """
        Helper function to get values from keys, where keys may not exist

        :param fixture: The fixture dictionary of data that parsed data should match
        :param key1: The section of the fixture (client / os / device)
        :param key2: The key of the value to test
        """
        if not fixture:
            return ''

        if key1 not in fixture:
            return ''

        if not fixture[key1] or key2 not in fixture[key1]:
            return ''

        return str(fixture[key1][key2])

    def test_parsing(self):

        for fixture in self.load_fixtures():
            self.user_agent = unquote(fixture.pop('user_agent'))
            device = DeviceDetector(self.user_agent, headers=fixture.get('headers', {}))
            device.parse()

            # OS properties
            self.assertEqual(
                self.get_value(fixture, 'os', 'name'), device.os_name(), field='os_name'
            )
            self.assertEqual(
                self.get_value(fixture, 'os', 'version'), device.os_version(), field='os_version'
            )

            # Client properties
            parsed_name = device.client_name()
            name = APPID_TO_NAME.get(parsed_name, parsed_name)
            self.confirm_client_name(fixture, parsed_name=name)
            self.confirm_client_type(fixture, parsed_value=device.client_type())
            self.confirm_version(fixture, parsed_value=device.client_version())

            # Device properties
            self.matches_fixture_or_generic(
                fixture, key1='device', key2='type', parsed_value=device.device_type()
            )
            self.matches_fixture_or_generic(
                fixture, key1='device', key2='model', parsed_value=device.device_model()
            )
            self.matches_fixture_or_generic(
                fixture, key1='device', key2='brand', parsed_value=device.device_brand()
            )


class ParserBaseTest(Base):

    fixture_files = []
    fixture_key = 'client'  # key of fixture dict containing the values to compare
    fields = []
    Parser = None

    def load_fixtures(self):
        fixtures = []
        for ffile in self.fixture_files:
            with open(f'{ROOT}/{ffile}', 'r') as r:
                fixtures.extend(yaml.load(r, SafeLoader))
        return fixtures

    def test_parsing(self):
        if not self.Parser or not self.fields:
            return

        fixtures = self.load_fixtures()

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            expect = fixture[self.fixture_key]
            parsed = self.Parser(
                self.user_agent,
                client_hints=ClientHints.new(fixture.get('headers', {})),
            ).clear_cache().parse()  # clear cache because fixture files may contain duplicate UAs
            data = parsed.ua_data

            for field in self.fields:
                self.assertIn(
                    field,
                    data,
                    msg=f'Error parsing {self.user_agent!r}. Parsed data missing {field!r} key.'
                )
                self.assertEqual(
                    str(expect.get(field, '')),
                    str(data.get(field, '')),
                    field=field,
                )


class GenericParserTest(ParserBaseTest):

    skipped = []

    def test_skipped_useragents(self):
        for ua in self.skipped:
            parsed = self.Parser(ua, None).parse()
            self.assertEqual(parsed.ua_data, {})


if __name__ == '__main__':
    unittest.main()



from . import BaseClientParser
from ua_extract.enums import AppType


class MediaPlayer(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.MediaPlayer

    fixture_files = [
        'local/client/mediaplayers.yml',
        'upstream/client/mediaplayers.yml',
    ]


__all__ = [
    'MediaPlayer',
]



from typing import TYPE_CHECKING

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self
from .lazy_regex import RegexLazy
from .enums import DeviceType

from .parser import (  # type: ignore[attr-defined]
    BaseClientParser,
    BaseDeviceParser,
    ClientHints,
    OS,
    # Device extractors
    Bot,
    Camera,
    CarBrowser,
    Console,
    Device,
    HbbTv,
    PortableMediaPlayer,
    Notebook,
    ShellTv,
    MOBILE_DEVICE_TYPES,
    # Clients
    AdobeCC,
    DictUA,
    Browser,
    FeedReader,
    Library,
    MediaPlayer,
    Messaging,
    MobileApp,
    OsUtility,
    Antivirus,
    DesktopApp,
    PIM,
    VPNProxy,
    # Generic name extractors
    ApplicationIDExtractor,
    NameVersionExtractor,
    WholeNameExtractor,
)
from .parser.settings import APPLE_OS_NAMES, TV_CLIENTS
from .settings import BOUNDED_REGEX, DDCache, WORTHLESS_UA_TYPES
from .utils import (
    clean_ua,
    long_ua_no_punctuation,
    mostly_numerals,
    mostly_repeating_characters,
    only_numerals_and_punctuation,
    random_alphanumeric_string,
    uuid_like_name,
    ua_hash_key,
)
from .yaml_loader import normalized_regex_list

DESKTOP_FRAGMENT = RegexLazy(BOUNDED_REGEX.format(r'(?:Windows (?:NT|IoT)|X11; Linux x86_64)'))


class DeviceDetector:
    if TYPE_CHECKING:
        parsed: bool

    fixture_files = [
        'local/device/normalize.yml',
    ]

    CLIENT_PARSERS = (
        AdobeCC,
        DictUA,
        FeedReader,
        Messaging,
        MobileApp,
        MediaPlayer,
        PIM,
        VPNProxy,
        OsUtility,
        Antivirus,
        DesktopApp,
        Browser,
        Library,
        NameVersionExtractor,
        WholeNameExtractor,
    )

    # The order needs to be the same as the order of device
    # parser classes used in the matomo project
    DEVICE_PARSERS = (
        HbbTv,
        ShellTv,
        Console,
        CarBrowser,
        Camera,
        PortableMediaPlayer,
        Notebook,
        Device,
    )

    __slots__ = (
        'client',
        'device',
        'model',
        'user_agent_lower',
        'user_agent',
        'ua_hash',
        'bot',
        'os',
        'skip_bot_detection',
        'skip_device_detection',
        'all_details',
        'parsed',
        'headers',
        'client_hints',
        '_normalized_regex_list',
    )

    def __new__(
        cls,
        user_agent: str,
        skip_bot_detection: bool = False,
        skip_device_detection: bool = False,
        headers: dict[str, str] | None = None,
    ) -> 'DeviceDetector':
        ua_key = f'{user_agent.lower()}-{skip_bot_detection}-{skip_device_detection}'
        uah = ua_hash_key(ua_key, headers)
        if cached := DDCache['user_agents'].get(uah):
            cached.parsed = True
            return cached

        res = super().__new__(cls)
        res.ua_hash = uah
        return res

    def __init__(
        self,
        user_agent: str,
        skip_bot_detection: bool = False,
        skip_device_detection: bool = False,
        headers: dict[str, str] | None = None,
    ) -> None:
        """

        Args:
            user_agent: User Agent string to parse
            skip_bot_detection: Skip checking if client is a bot
            skip_device_detection: Skip device brand and model lookup.
            headers: Client Hint headers from the request
        """
        # Prevent reinitialization of memoized classes
        if getattr(self, 'parsed', False):
            return

        # Holds the useragent that should be parsed
        self.user_agent_lower = user_agent.lower()
        self.user_agent = clean_ua(user_agent, self.user_agent_lower)
        ua_key = f'{self.user_agent_lower}-{skip_bot_detection}-{skip_device_detection}'
        self.ua_hash = ua_hash_key(ua_key, headers)
        self.os: OS | None = None
        self.client: BaseClientParser | None = None
        self.device: BaseDeviceParser | None = None
        self.model = ''

        # Holds bot information if parsing the UA results in a bot
        # (All other information attributes will stay empty in that case)
        # If self.discardBotInformation is set to True, this property will be set to
        # True if parsed UA is identified as bot, additional information will be not available
        # If self.skip_bot_detection is set to True, bot detection will not be performed and
        # isBot will always be False
        self.bot: Bot | None = None

        self.skip_bot_detection = skip_bot_detection
        self.skip_device_detection = skip_device_detection
        self.all_details: dict = {'normalized': ''}  # type: ignore[type-arg]
        self.headers = headers or {}
        self.client_hints = ClientHints.new(headers) if headers else None
        self._normalized_regex_list = normalized_regex_list(self.fixture_files)

    @property
    def class_name(self) -> str:
        return self.__class__.__name__

    # -----------------------------------------------------------------------------
    # UA parsing methods
    # -----------------------------------------------------------------------------
    def is_digit(self) -> bool:
        """
        Remove all punctuation. If only digits remain,
        don't bother saving, as nothing can be learned.

        21/4.35.1.2
        5.0.6

        Or if entire string is mostly numeric, discard
        15B93
        """
        if only_numerals_and_punctuation(self.user_agent):
            return True

        return mostly_numerals(self.user_agent)

    def is_uuid(self) -> bool:
        """
        Check for strings that are UUIDs

        (UUIDs enclosed in curly braces are valid)
        {1378F00B-BCEA-418F-B1AF-C343EA4F9417}
        {022CCD4D-EDE3-40EB-BE1D-FE4041B0A050}

        A:08338459-4ca1-457f-a596-94c3a9037d20
        I:5DFF6AEC-DCED-4BA0-B122-B1826C1CEB02
        """
        ua = self.user_agent.strip('({})')
        if len(ua) >= 2 and ua[1] == ':':
            ua = self.user_agent[2:]

        return uuid_like_name(ua)

    def is_gibberish(self) -> bool:
        """
        Check for frequently occurring patterns of meaninglessness
        """
        if mostly_repeating_characters(self.user_agent_lower):
            return True

        if random_alphanumeric_string(self.user_agent_lower):
            return True

        return long_ua_no_punctuation(self.user_agent_lower)

    def normalize(self) -> str:
        """
        Check for common worthless features that preclude the need for any further processing.

        If UA string is not worthless, process against normalizing regexes.

        Some client User Agent strings such as Antivirus services add file version information
        that is of no use outside the application itself. Remove such information to present a
        cleaner UA string with fewer duplicates
        """
        if normalized := self.all_details.get('normalized'):
            return normalized

        if self.is_digit():
            self.all_details['normalized'] = 'Numeric'
        elif self.is_uuid():
            self.all_details['normalized'] = 'UUID'
        elif self.is_gibberish():
            self.all_details['normalized'] = 'Gibberish'
        else:
            for nr in self._normalized_regex_list:
                regex = nr['regex']
                groups = r'{}'.format(nr['groups'])
                ua = regex.sub(groups, self.user_agent)
                if ua != self.user_agent:
                    self.all_details['normalized'] = ua
                    break
            else:
                self.all_details['normalized'] = ''

        return self.all_details['normalized']

    def is_worthless(self) -> bool:
        """
        Is this UA string of no possible interest?
        """
        self.normalize()
        if self.headers:
            return False
        return self.all_details['normalized'] in WORTHLESS_UA_TYPES

    def parse(self) -> Self:
        if cached := DDCache['user_agents'].get(self.ua_hash):
            return cached

        if not self.user_agent and not self.headers:
            return self

        if not self.skip_bot_detection:
            self.parse_bot()

        if self.is_worthless():
            return self

        self.parse_os()

        self.parse_client()

        if not self.skip_device_detection:
            self.parse_device()
            # All devices running Coolita OS are assumed to be a tv
            if self.os_name() == 'Coolita OS':
                device_data = {
                    'brand': 'coocaa',
                    'type': DeviceType.TV,
                }
                try:
                    self.all_details['device'] |= device_data
                except KeyError:
                    self.all_details['device'] = device_data

        self.parsed = True
        DDCache['user_agents'][self.ua_hash] = self
        return self

    def supplement_secondary_client_data(self, app_idx: ApplicationIDExtractor) -> None:
        """
        Add data to secondary_client details
        """
        if not self.client:
            return
        self.client.secondary_client.update(app_idx.details)
        try:
            self.all_details['client']['secondary_client'] |= app_idx.details
        except KeyError:
            self.all_details['client']['secondary_client'] = app_idx.details

    def parse_client(self) -> None:
        """
        Parses the UA for client information using the Client parsers
        """
        if self.client:
            return None

        os_details = self.all_details.get('os', {})
        for Parser in self.CLIENT_PARSERS:
            parser = Parser(
                self.user_agent,
                self.client_hints,
                os_details=os_details,
            ).parse()

            if parser.ua_data:
                self.client = parser
                self.all_details['client'] = parser.ua_data
                break

        return self.extract_app_id()

    def extract_app_id(self) -> None:
        """
        Extract app_id from UA if not found in client hints.
        """
        if self.client and not self.client.CHECK_APP_ID:
            return

        # If app_id already present, it was extracted from client hints
        if self.all_details.get('client', {}).get('app_id'):
            return

        app_idx = ApplicationIDExtractor(self.user_agent).extract()

        if app_idx.details.get('app_id', ''):
            if not self.all_details.get('client'):
                self.all_details['client'] = app_idx.details
                return

            self.supplement_secondary_client_data(app_idx)

    def parse_device(self) -> None:
        """
        Parses the UA for Device information using the Device or Bot parsers
        """
        if self.device or self.skip_device_detection:
            return

        os_details = self.all_details.get('os', {})

        for Parser in self.DEVICE_PARSERS:
            parser = Parser(
                self.user_agent,
                self.client_hints,
                os_details=os_details,
            ).parse()
            if parser.ua_data:
                self.device = parser
                self.all_details['device'] = parser.ua_data

                if self.is_television():
                    self.all_details['device']['type'] = DeviceType.TV
                return

    def parse_bot(self) -> None:
        """
        Parses the UA for bot information using the Bot parser
        """
        if not self.skip_bot_detection and not self.bot:
            self.bot = Bot(self.user_agent, self.client_hints).parse()
            self.all_details['bot'] = self.bot.ua_data

    def parse_os(self) -> None:
        """
        Parses the UA for Operating System information using the OS parser
        """
        if not self.os:
            os = OS(self.user_agent, self.client_hints).parse()
            if os:
                self.os = os
                self.all_details['os'] = os.ua_data

    # -----------------------------------------------------------------------------
    # Data post-processing / analysis
    # -----------------------------------------------------------------------------
    def is_known(self) -> bool:
        for section, data in self.all_details.items():
            if data:
                return True
        return False

    def is_bot(self) -> bool:
        return bool(self.all_details.get('bot'))

    def bot_name(self) -> str:
        return self.all_details.get('bot', {}).get('name', '')

    def is_television(self) -> bool:
        """
        Detect devices that are likely TVs.

        All devices that contain Andr0id in string are assumed to be a tv
        All devices running Tizen TV or SmartTV are assumed to be a tv
        Devices running known tv clients are assumed to be a TV
        All devices containing TV fragment are assumed to be a tv
        All devices running Coolita OS are assumed to be a tv
        """
        if self.client_name() in TV_CLIENTS:
            return True

        return self.device_type() == DeviceType.TV

    def uses_mobile_browser(self) -> bool:
        if isinstance(self.client, Browser):
            return self.client.is_mobile_only()
        return False

    def engine(self) -> str:
        if 'browser' not in self.client_type():
            return ''
        return self.all_details.get('client', {}).get('engine') or ''

    def is_mobile(self) -> bool:
        """
        Returns if the parsed UA is detected as a mobile device
        """
        if self.client_hints and self.client_hints.mobile:
            return self.client_hints.is_mobile()

        if self.device_type() in MOBILE_DEVICE_TYPES:
            return True
        return not self.is_bot() and not self.is_desktop() and not self.is_television()

    def is_desktop(self) -> bool:
        """
        Returns if the parsed UA was identified as desktop device
        Desktop devices are all devices with an unknown type that are running a desktop os
        """
        if self.client_hints and self.client_hints.mobile:
            return self.client_hints.is_desktop()

        if self.uses_mobile_browser():
            return False

        return self.device_type() == DeviceType.Desktop

    def is_feature_phone(self) -> bool:
        """
        Check for various indicators that this is feature phone.
        """
        return self.device_type() == DeviceType.FeaturePhone

    def client_name(self) -> str:
        return self.all_details.get('client', {}).get('name') or ''

    def client_version(self) -> str:
        return self.all_details.get('client', {}).get('version') or ''

    def client_application_id(self) -> str:
        """
        Return Apple Bundle ID or Android Package ID if present.

        2ndLine/4.8.1 (com.second.phonenumber; build:1.5; iOS 16.7.12) Alamofire/5.4.4
        AcuityApp/5.14.0 (com.acuityscheduling.app.ios; build:1686757700; iPhone; iOS 17.1.1) SquarespaceMobileiOS
        """
        client = self.all_details.get('client', {})
        return client.get('app_id', '') or client.get('secondary_client', {}).get('app_id') or ''

    def client_type(self) -> str:
        return self.all_details.get('client', {}).get('type') or ''

    def secondary_client_name(self) -> str:
        return self.all_details.get('client', {}).get('secondary_client', {}).get('name') or ''

    def secondary_client_version(self) -> str:
        return self.all_details.get('client', {}).get('secondary_client', {}).get('version') or ''

    def secondary_client_type(self) -> str:
        return self.all_details.get('client', {}).get('secondary_client', {}).get('type') or ''

    def preferred_client_name(self) -> str:
        """
        Android and iOS mobile browsers often contain more interesting
        app information.

        If Secondary app can be extracted, prefer those app details
        as being more specific.
        """
        return self.secondary_client_name() or self.client_name() or self.client_application_id()

    def preferred_client_version(self) -> str:
        return self.secondary_client_version() or self.client_version()

    def preferred_client_type(self) -> str:
        return self.secondary_client_type() or self.client_type()

    def device_type(self) -> DeviceType:
        """
        Get device type, preferably from the Device Parser, but
        calculate from other attributes Device Parser failed.

        Should work, even if skip_device_detection=True
        """
        return self.all_details.get('device', {}).get('type') or DeviceType.Unknown

    def device_model(self) -> str:
        """
        Detect model from UserAgent, and fall back to checking Client Hints
        """
        client_hints_model = self.client_hints and self.client_hints.model or ''
        if self.skip_device_detection:
            return client_hints_model
        return self.all_details.get('device', {}).get('model') or client_hints_model

    def device_brand(self) -> str:
        if self.skip_device_detection:
            return ''

        if brand := self.all_details.get('device', {}).get('brand'):
            return brand

        # Assume all devices running iOS / macOS are from Apple
        if self.os_name() in APPLE_OS_NAMES:
            return 'Apple'

        return ''

    def os_name(self) -> str:
        return self.all_details.get('os', {}).get('name') or ''

    def os_version(self) -> str:
        return self.all_details.get('os', {}).get('version') or ''

    def pretty_name(self) -> str:
        return self.all_details.get('normalized') or self.user_agent or ''

    def pretty_print(self) -> str:
        if not self.is_known():
            return self.user_agent
        os = client = device = 'N/A'
        if self.os_name():
            os = f'{self.os_name()} {self.os_version()}'
        if self.client_name():
            client = f'{self.client_name()} {self.client_version()} ({self.client_type().title()})'
        if self.device_model():
            device = f'{self.device_model()} ({self.device_type().title()})'
        return f'Client: {client} Device: {device} OS: {os}'.strip()

    def __str__(self) -> str:
        return self.user_agent

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.user_agent!r})'


class SoftwareDetector(DeviceDetector):
    def __init__(
        self,
        user_agent: str,
        skip_bot_detection: bool = True,
        skip_device_detection: bool = True,
        headers: dict[str, str] | None = None,
    ):
        super().__init__(
            user_agent,
            skip_bot_detection=skip_bot_detection,
            skip_device_detection=skip_device_detection,
            headers=headers,
        )


__all__ = (
    'DeviceDetector',
    'SoftwareDetector',
)



from .base import *
from .bot import *
from .camera import *
from .car_browser import *
from .console import *
from .device import *
from .notebook import *
from .portable_media_player import *
from .television import *
from .vendor_fragment import *



from urllib.parse import unquote
from ..base import ParserBaseTest
from ...parser import (
    ApplicationIDExtractor,
)


class TestApplicationIDExtractor(ParserBaseTest):

    fixture_files = [
        'tests/parser/fixtures/local/extractor/applicationid.yml',
        'tests/parser/fixtures/local/extractor/app_id_override_name.yml',
    ]

    def test_parsing(self):
        fixtures = self.load_fixtures()
        error = 'Error parsing {}.\n Parsed value "{}" != expected value "{}"'

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            app_id = ApplicationIDExtractor(self.user_agent).extract()

            expected = fixture['client']['pretty_name']
            parsed = app_id.pretty_name()
            self.assertEqual(expected, parsed, msg=error.format(self.user_agent, parsed, expected))

__all__ = [
    'TestApplicationIDExtractor',
]



import string

from ...enums import AppType
from ...lazy_regex import RegexLazyIgnore
from ..parser import Parser
from ...utils import calculate_dtype

keep = frozenset(['!', '@', '+'])
table = str.maketrans(dict.fromkeys(''.join(c for c in string.punctuation if c not in keep)))

# fmt: off
UNWANTED_UA_STRINGS = [
    # 13F7BD1A-F6FF-411E-BF5E
    # 4UWWU-WBDXC-VYFN3-QDJMH
    RegexLazyIgnore(r"([\d\w]{4,10}-){3}[\d\w]{4,10}$"),

    # long gibberish string, starting with integer and containing no punctuation
    RegexLazyIgnore(r'^\d[\d\w]{20,40}$'),
]

UNWANTED_APP_NAMES = [
    RegexLazyIgnore(r'sm-\w+-android'),
    RegexLazyIgnore(r'^4d531b'),

    # App IDs will be parsed with ApplicationID extractor
    RegexLazyIgnore(r'^com\.'),
]
# fmt: on


class BaseClientParser(Parser):
    # If this client parser class matches, also
    # check the UA string for an application id.
    # Disable that check if the application id
    # is used for many different apps and would
    # result in a loss of precise app names.
    CHECK_APP_ID = True
    __slots__ = ()
    APP_TYPE = AppType.Unknown

    def set_data_from_client_hints(self) -> None:
        """
        Prefer client hints over user agent data generally.
        Override on subclasses to customize order
        """
        if self.ua_data and self.ch_client_data:
            self.ua_data |= self.ch_client_data

    def dtype(self) -> AppType | str:
        return self.APP_TYPE

    def set_details(self) -> None:
        """
        Set app data from UA or Client Hints.
        """
        self.set_data_from_client_hints()

        super().set_details()
        if self.known and not self.ua_data.get('type'):
            self.ua_data['type'] = self.dtype()

        if custom_app_details := self.appdetails_data.get(self.name().lower()):
            self.ua_data |= custom_app_details


class GenericClientParser(BaseClientParser):
    APP_TYPE = AppType.Generic

    # -------------------------------------------------------------------
    # App names that have no value to us so we want to discard them
    # Should be lowercase
    discard = {
        'productname',
        'null',
        'httppostlib',
        'mozilla',
        'mobileios',
    }

    # -------------------------------------------------------------------
    # List of substrings that if found in the app name, we will
    # discard the entire app name
    # Should be lowercase
    unwanted_substrings = [
        'ab_1.1.3011',
        'deviceid=',
        'timezone=',
    ]

    def discard_name(self) -> bool:
        """
        Determine if app name is of any value to us

        Return True if app should be discarded
        """
        if self.unwanted_regex_match():
            return True

        if not self.is_name_length_valid():
            return True

        if self.app_name_no_punc().lower() in self.discard:
            return True

        if self.is_substring_unwanted():
            return True

        if self.unwanted_regex_name_match():
            return True

        return self.is_name_mostly_numeric()

    def is_name_length_valid(self) -> bool:
        """
        Check if app name portion of UA is between 1 and 45 chars
        """
        return 1 < len(self.app_name) <= 45

    def is_substring_unwanted(self) -> bool:
        for substring in self.unwanted_substrings:
            if substring in self.app_name.lower():
                return True

        return False

    def unwanted_regex_match(self) -> bool:
        for regex in UNWANTED_UA_STRINGS:
            if regex.search(self.user_agent):
                return True

        return False

    def unwanted_regex_name_match(self) -> bool:
        for regex in UNWANTED_APP_NAMES:
            if regex.search(self.app_name):
                return True

        return False

    def is_name_mostly_numeric(self) -> bool:
        """
        Strip punctuation from app name and return True if
        alphabetic characters are less than 50% of the string
        """
        app_no_punc = self.app_name_no_punc()
        if not app_no_punc:
            return True

        try:
            int(app_no_punc)
            return True
        except ValueError:
            pass

        alphabetic_chars = 0
        for char in app_no_punc:
            if not char.isnumeric():
                alphabetic_chars += 1

        threshold = 0.25 if len(self.app_name) < 10 else 0.5

        return alphabetic_chars / len(app_no_punc) < threshold

    def app_name_no_punc(self) -> str:
        """
        Remove punctuation from the given string and return
        the new string
        """
        if self.app_name_no_punctuation:
            return self.app_name_no_punctuation

        self.app_name_no_punctuation = self.app_name.translate(table)

        return self.app_name_no_punctuation

    def dtype(self) -> AppType | str:
        return calculate_dtype(app_name=self.app_name) or self.APP_TYPE


__all__ = (
    'BaseClientParser',
    'GenericClientParser',
)



from . import BaseClientParser
from ua_extract.enums import AppType


class FeedReader(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.FeedReader

    fixture_files = [
        'upstream/client/feed_readers.yml',
    ]


__all__ = [
    'FeedReader',
]



import unittest

from ...parser.key_value_pairs import key_value_pairs


class TestKeyValuePairs(unittest.TestCase):

    def test_trailing_key_space_version(self):
        ua = 'Mozilla/5.0 (Linux; Android 8.0.0; SM-A530W Build/R16NW; wv) AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/73.0.3683.90 Mobile Safari/537.36 Android SermonAudio.com 1.9.8'

    def test_2(self):
        ua = 'YouCam Fun/22261949 (iPhone; iOS 12.0.1; Scale/2.00)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('youcamfun', 'YouCam Fun', '22261949')],
        )

    def test_3(self):
        ua = 'Aurora HDR 2018/1.1.2 Sparkle/1.13.1'
        self.assertListEqual(
            key_value_pairs(ua),
            [('aurorahdr2018', 'Aurora HDR 2018', '1.1.2')],
        )

    def test_4(self):
        ua = 'libreoffice 5.4.3.2 (92a7159f7e4af62137622921e809f8546db437e5; windows; x86;)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('libreoffice', 'libreoffice', '5.4.3.2')],
        )

    def test_5(self):
        ua = 'Microsoft Office Access 2013 (15.0.4693) Windows NT 6.2'
        self.assertListEqual(
            key_value_pairs(ua),
            [('microsoftofficeaccess', 'Microsoft Office Access', '2013')],
        )

    def test_with_url1(self):
        ua = 'DigiCal (v1.8.2b; http://digibites.nl/digical)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('digical', 'DigiCal', '1.8.2')],
        )

    def test_with_url2(self):
        ua = 'Mozilla/5.0 (Stellarium Bright Novae Plugin 0.1.3; http://stellarium.org/)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('stellariumbrightnovaeplugin', 'Stellarium Bright Novae Plugin', '0.1.3')],
        )

    def test_6(self):
        ua = 'The Economist on iPad NA/69 CFNetwork/893.14.2 Darwin/17.3.0'
        self.assertListEqual(
            key_value_pairs(ua),
            [('theeconomistonipadna', 'The Economist on iPad NA', '69')],
        )

    def test_7(self):
        ua = 'The Columbia Bank, MD/5.11.0.42 BundleID/com.intuit.mobilebanking01368 BundleDeviceFamily/iPhone,iPad (iPhone; iPhone11,2; iPhone XS; iOS 12.1.4)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('thecolumbiabank,md', 'The Columbia Bank, MD', '5.11.0.42'), ('bundleid', 'BundleID', 'com.intuit.mobilebanking01368'), ('bundledevicefamily', 'BundleDeviceFamily', 'iPhone')],
        )

    def test_8(self):
        ua = 'Podbean/iOS (http://podbean.com) 5.0 - ba7575612c9a5554c0b8f5ffc378d1c7'
        self.assertListEqual(
            key_value_pairs(ua),
            [('podbean', 'Podbean', 'iOS')],
        )

    def test_9(self):
        ua = 'wsdl2objc iPhone SermonAudio.com 4.4.5'
        self.assertListEqual(
            key_value_pairs(ua),
            [('wsdl2objciphonesermonaudio.com', 'wsdl2objc iPhone SermonAudio.com', '4.4.5')],
        )

    def test_10(self):
        ua = 'AAM 1.1 rv:1.0 (iPhone; iOS 12.0; en_US)'
        self.assertListEqual(
            key_value_pairs(ua),
            [('aam', 'AAM', '1.1')],
        )

    def test_11(self):
        ua = 'Microsoft URL Control - 6.01.9782'
        self.assertListEqual(
            key_value_pairs(ua),
            [('microsofturlcontrol', 'Microsoft URL Control', '6.01.9782')],
        )

    def test_12(self):
        ua = '1.0,win/10.0.16299,AV/17.1.2286,avl/devcontrol/17.1.3394.0,ffl'
        self.assertListEqual(
            key_value_pairs(ua),
            [('av', 'AV', '17.1.2286'), ('avl', 'avl', 'devcontrol')],
        )

    def test_13(self):
        ua = 'ArcGISiOS-10.2.5/10.3.2/iPad6,12'
        self.assertListEqual(
            key_value_pairs(ua),
            [('arcgisios-10.2.5', 'ArcGISiOS-10.2.5', '10.3.2')],
        )

    def test_is_browser_nothing_interesting1(self):
        ua = 'mozilla/5.0 (windows nt 10.0; win64; x64) applewebkit/537.36 (khtml, like gecko) chrome/46.0.2486.0 safari/537.36 edge/13.10586 edge 13.0'
        self.assertListEqual(
            key_value_pairs(ua),
            [],
        )

    def test_is_browser_nothing_interesting2(self):
        ua = 'Safari/14607.1.39 CFNetwork/978.0.6 Darwin/18.5.0 (x86_64)'
        self.assertListEqual(
            key_value_pairs(ua),
            [],
        )

    def test_is_browser_nothing_interesting3(self):
        ua = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/73.0.3683.88 Safari/537.36 Vivaldi/2.4.1488.35'
        self.assertListEqual(
            key_value_pairs(ua),
            [],
        )

    def test_is_browser_nothing_interesting4(self):
        ua = 'Vivaldi/2.0.1309.37 WinSparkle/0.5.2 (Win64)'
        self.assertListEqual(
            key_value_pairs(ua),
            [],
        )

    def test_is_browser_nothing_interesting5(self):
        ua = 'Opera/9.62 (J2ME/MIDP; Opera Mini/5.1.37933421/28.62; U; en) Presto/2.521.26 Version/11.521'
        self.assertListEqual(
            key_value_pairs(ua),
            [('version', 'Version', '11.521')],
        )

    def test_is_browser_nothing_interesting6(self):
        ua = 'Opera/9.62 (Android 4.1.2; Linux; Opera Mobi/ADR-25672775) Presto/2.520.13 Version/12.520'
        self.assertListEqual(
            key_value_pairs(ua),
            [('version', 'Version', '12.520')],
        )




from .base import BaseDeviceParser
from ua_extract.enums import DeviceType


class PortableMediaPlayer(BaseDeviceParser):
    __slots__ = ()
    DEVICE_TYPE = DeviceType.PortableMediaPlayer

    fixture_files = [
        'upstream/device/portable_media_player.yml',
    ]


__all__ = [
    'PortableMediaPlayer',
]



from . import BaseClientParser
from ua_extract.enums import AppType


class VPNProxy(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.VpnProxy

    fixture_files = [
        'local/client/vpnproxy.yml',
    ]


__all__ = [
    'VPNProxy',
]



from urllib.parse import unquote
from ..base import ParserBaseTest
from ...parser import Camera, CarBrowser, Console, Notebook


class TestDeviceBase(ParserBaseTest):

    def test_parse(self):
        fixtures = self.load_fixtures()

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            expect = fixture['device']
            parsed = self.Parser(self.user_agent, None).parse()

            data = parsed.ua_data

            for field in ('type', 'brand', 'model'):
                if field not in data:
                    continue
                self.assertEqual(
                    str(expect[field]),
                    str(data[field]),
                    field=field,
                )


class TestConsole(TestDeviceBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/device/console.yml',
    ]
    Parser = Console


class TestCamera(TestDeviceBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/device/camera.yml',
    ]
    Parser = Camera


class TestCarBrowser(TestDeviceBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/device/car_browser.yml',
    ]
    Parser = CarBrowser


class TestNotebook(TestDeviceBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/device/notebook.yml',
    ]
    Parser = Notebook


__all__ = [
    'TestConsole',
    'TestCamera',
    'TestCarBrowser',
    'TestNotebook',
]



from collections import defaultdict
from typing import Any, TypedDict
import yaml
import ahocorasick_rs
from pathlib import Path
import warnings

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader  # type: ignore[assignment]

import ua_extract
from .lazy_regex import RegexLazyIgnore
from .settings import BOUNDED_REGEX, DDCache, ROOT
from .enums import AppType


class RegexLoader:
    # Paths to yml files of regexes
    fixture_files: list[str] | tuple[str, ...] = ()

    # Constant used as value for unknown browser / os
    UNKNOWN = 'UNK'

    @property
    def cache_name(self) -> str:
        """Class name, used for cache key"""
        return self.__class__.__name__

    @staticmethod
    def load_from_yaml(yfile: str) -> dict[str, Any] | list[dict[str, Any]]:
        """
        Load yaml from regexes directory, or extract from the egg
        """
        yml_file_path = f'{ROOT}/{yfile}'
        if Path(yml_file_path).exists():
            with open(yml_file_path, 'r', encoding="utf-8") as yf:
                return yaml.load(yf, SafeLoader)

        try:
            yfile = f'ua_extract/{yfile}'
            return yaml.load(ua_extract.__loader__.get_data(yfile), SafeLoader)  # type: ignore[union-attr]
        except OSError:
            if "upstream" in yfile:
                warnings.warn(
                    f"File {yfile} not found in regexes directory. Try updating the user agents.",
                    category=UserWarning,
                    stacklevel=2,
                )

            return {}

    def yaml_to_list(self, yfile: str) -> list[dict[str, Any]]:
        """
        Override method on subclasses if yaml format varies.

        Load yaml file to list of dicts
        """
        regexes = self.load_from_yaml(yfile)
        if isinstance(regexes, list):
            return regexes

        reg_list = []
        for entry, regex in regexes.items():
            regexes[entry]['brand'] = entry
            reg_list.append(regex)

        return reg_list

    @property
    def regex_list(self) -> list[dict[str, Any]]:
        if (regexes := DDCache['regexes'].get(self.cache_name)) is not None:
            return regexes

        all_regexes = []
        for fixture in self.fixture_files:
            regexes = self.yaml_to_list(f'regexes/{fixture}')

            for regex in regexes:
                if 'regex' in regex:
                    regex['regex'] = RegexLazyIgnore(BOUNDED_REGEX.format(regex['regex']))
                for model in regex.get('models', []):
                    model['regex'] = RegexLazyIgnore(BOUNDED_REGEX.format(model['regex']))
                for version in regex.get('versions', []):
                    version['regex'] = RegexLazyIgnore(BOUNDED_REGEX.format(version['regex']))

            all_regexes.extend(regexes)

        DDCache['regexes'][self.cache_name] = all_regexes

        return all_regexes

    def load_ahocorasick_patterns(self) -> ahocorasick_rs.AhoCorasick | None:
        """
        Load AhoCorasick words from file, or expand from regexes.
        """
        try:
            return DDCache['corasick'][self.cache_name]
        except KeyError:
            pass

        all_corasick_words: set[str] = set()
        for fixture in self.fixture_files:
            ac_fixture = f'regexes/ahocorasick/{fixture}'
            manual = self.load_manually_defined_words()
            all_corasick_words.update(set(manual.get('Words') or set()))

            if words := set(self.load_from_yaml(ac_fixture)):
                all_corasick_words.update(words)  # type: ignore[arg-type]

        ac = ahocorasick_rs.AhoCorasick(all_corasick_words) if all_corasick_words else None
        DDCache['corasick'][self.cache_name] = ac

        return ac

    def load_manually_defined_words(self) -> dict[str, list[str]]:
        """
        Every Parser or Detector class can have a set of words
        and Word exclusions that are manually defined.
        """
        return self.load_from_yaml(f'regexes/ahocorasick/classes/{self.cache_name}.yml') or {}  # type: ignore[return-value]

    def clear_cache(self) -> Self:
        """
        Helper method to clear cache on tests.
        """
        DDCache.clear_user_agents()
        return self


class AppNameType(TypedDict):
    name: str
    type: AppType | str


def app_pretty_names_types_data() -> dict[str, AppNameType]:
    """
    Load App Details data into dictionary.

    General regex extracts all name/version entries of interest from the UA
    string, and each ParserClass will check to see if any of those names is
    contained in the relevant appdetails.yml file. Much faster than writing
    individual regexes for each app.
    """
    cache_key = 'app_details'
    appdetails = DDCache.get(cache_key) or {}
    if appdetails:
        return appdetails

    regex_loader = RegexLoader()
    all_app_details = {}
    for fixture, dtype in (
        ('appdetails/ai.yml', AppType.ArtificialIntelligence),
        ('appdetails/av.yml', AppType.Antivirus),
        ('appdetails/browser.yml', AppType.Browser),
        ('appdetails/desktop_app.yml', AppType.DesktopApp),
        ('appdetails/game.yml', AppType.Game),
        ('appdetails/library.yml', AppType.Library),
        ('appdetails/mediaplayer.yml', AppType.MediaPlayer),
        ('appdetails/messaging.yml', AppType.Messaging),
        ('appdetails/mobile_app.yml', AppType.MobileApp),
        ('appdetails/navigation.yml', AppType.Navigation),
        ('appdetails/osutility.yml', AppType.OsUtility),
        ('appdetails/p2p.yml', AppType.P2P),
        ('appdetails/pim.yml', AppType.PIM),
        ('appdetails/productivity.yml', AppType.Productivity),
        ('appdetails/vpnproxy.yml', AppType.VpnProxy),
    ):
        all_app_details[dtype] = regex_loader.yaml_to_list(fixture)

    # convert uaname value to dict key and remove spaces and add that key as well.
    generalized_details: dict = defaultdict(dict)  # type: ignore[type-arg]
    for dtype, entries in all_app_details.items():
        for entry in entries:
            name = entry['name']
            key = entry['uaname'].lower().replace(' ', '')
            data = {
                'name': name,
                'type': entry.get('type', dtype),
            }
            generalized_details[key] = data

            # Match airmail, airmail-android, airmail-iphone
            suffixes = str(entry.get('suffixes', '')).lower().replace(' ', '')
            for suffix in suffixes.split('|'):
                if suffix:
                    generalized_details[f'{key}{suffix}'] = data

    # Load upstream fixtures last, so that any values that are added
    # upstream later will stomp on local changes.
    for fixture, dtype in (
        ('regexes/upstream/client/hints/browsers.yml', AppType.Browser),
        ('regexes/upstream/client/hints/apps.yml', AppType.MobileApp),
    ):
        for app_id, pretty_name in regex_loader.load_from_yaml(fixture).items():  # type: ignore[union-attr]
            generalized_details[app_id] = {
                'name': pretty_name,
                'type': dtype,
            }

    DDCache[cache_key] = generalized_details

    return generalized_details


def normalized_regex_list(fixture_files: list[str]) -> list[dict[str, Any]]:
    cache_key = 'normalize_regexes'
    regexes = DDCache.get(cache_key) or []
    if regexes:
        return regexes

    regex_loader = RegexLoader()
    for fixture in fixture_files:
        regexes.extend(regex_loader.yaml_to_list(f'regexes/{fixture}'))

    for regex in regexes:
        regex['regex'] = RegexLazyIgnore(regex['regex'])

    DDCache[cache_key] = regexes

    return regexes



from . import GenericClientParser

from ...lazy_regex import RegexLazyIgnore
from ..settings import SKIP_PREFIXES


# -------------------------------------------------------------------
# fmt: off
# Regexes that we use to parse UAs with a similar structure
parse_generic_regex = [
    # Weather_WeatherFoundation[1]_15E302
    # SpringBoard_WeatherFoundation[1]_16A
    (RegexLazyIgnore(r'(^(?:\w+)_WeatherFoundation).*'), 1),
    (RegexLazyIgnore(r'(^AVGSETUP).*'), 1),

    # ACC__14EDB170A0EAA78B40F
    (RegexLazyIgnore(r'(^ACC)__[A-Z0-9].*'), 1),

    # YWeather (iPhone/11.4.1; iOS; en_US;)
    # Mapbox iOS SDK (iPhone/11.1.1)
    # SCSDK/(iPhone;iOS;12.2)
    # Akamai NetSession C-API (win;AdDLMgr;capi_1.9.2;Vista)
    (RegexLazyIgnore(r'^([a-z0-9\- ]+)[ /]?\((?:iphone|ipad|win)'), 1),

    # CPIS&iPhone10.1&
    (RegexLazyIgnore(r'(^CPIS).*'), 1),

    # samsung SAMSUNG-SM-T337A SyncML_DM Client
    # samsung SMT377P SPDClient to samsung SM-T377P SPD-Client
    (RegexLazyIgnore(r'^samsung.*((?:SyncML_DM|SPD)[ \-]Client)$'), 1),
    (RegexLazyIgnore(r'(WXCommonUtils).*'), 1),

    # UBT_1EDA413E5BA1DEA76A
    # UBT_1ED
    (RegexLazyIgnore(r'^(UBT)_[\d\w]+$'), 1),

    # ACC10FDFFD20BCFEDA0D7D
    # ACC_4.00.3019_1BF78BFDF84E7EEFDA
    # ALU_1.02.3005_1EDDDB14FD0EBB93AE
    # ALU1023502106209EEF239E246AA
    # ALU__1023502106209EEF239E246AA
    (RegexLazyIgnore(r'^(ACC|ALU)_*[\d\w_\.]+$'), 1),

    # mShop::Telly_iPhone_13.7.0::iPad::iPhone_OS_ == Telly_iPhone_13.7.0
    # mShop::Amazon_Android_18.11.0.100::SAMSUNG-SM-G935A::Android_6.0.1
    # mShop::WindowShop_Android_16.13.0.850::SM-T817V::Android_6.0.1
    (RegexLazyIgnore(r'^mshop:+([a-z0-9_\.]+)'), 1),
]

# -------------------------------------------------------------------
# Regexes that we use to extract app versions
extract_version_regex = [
    # ANVSDKv.5.0.21
    # ANVSDKv5.0.21
    (RegexLazyIgnore(r'(v?[\.\d]+$)')),

    # Catch versions with trailing, separated characters
    # AppNamev.5.0.21_PRC = v.5.0.21
    (RegexLazyIgnore(r'((?:v.?)?\d[\d\.]+)')),
]
# fmt: on


class WholeNameExtractor(GenericClientParser):
    """
    Catch all for user agents that do not use the slash format
    """

    __slots__ = ()

    parse_generic_regex = parse_generic_regex
    extract_version_regex = extract_version_regex

    # -------------------------------------------------------------------
    def _parse(self) -> None:
        if self.ch_client_data:
            return

        self.clean_name()

        if self.discard_name() or not self.is_name_length_valid():
            return

        self.parse_name_version()

        # WholeNameExtractor is called to supply secondary app data
        # if the Browser class matches. So if only Browser data was
        # found then there is no "secondary" data!
        if self.app_name.lower() in SKIP_PREFIXES:
            return

        app_details = self.appdetails_data
        code = self.app_name.lower().replace(' ', '')

        try:
            self.app_name = app_details[code]['name']
        except KeyError:
            pass

        self.ua_data = {
            'name': self.app_name,
            'version': self.app_version,
            'type': app_details[code].get('type', ''),
        }

        self.known = True

    def clean_name(self) -> None:
        """
        Check if the extracted name uses a known format that we can
        extract helpful info from.  If so, update ua data and mark
        as known.
        """
        for regex, group in self.parse_generic_regex:
            m = regex.match(self.user_agent)

            if m:
                try:
                    self.app_name = m.group(group).strip()
                    return
                except Exception:
                    continue

        # Don't consider the entire UA as the name if it contains a slash.
        # In that case, there should be more parsing performed elsewhere.
        if '/' in self.user_agent:
            return

        self.app_name = self.user_agent

    def parse_name_version(self) -> str | None:
        """
        Check if app name has a suffix with the version number and
        extract if it does.  Return the UA string without the suffix.
        """
        for regex in self.extract_version_regex:
            match = regex.search(self.app_name)
            if match:
                self.app_version = match.group().strip()
                self.app_name = self.user_agent[: match.start()].strip(' /-')
                return self.app_version
        return None

    def is_name_length_valid(self) -> bool:
        """
        Check if app name portion of UA is between 3 and 60 chars
        """
        return 2 < len(self.app_name) <= 60



from collections import Counter
from hashlib import blake2s
from string import punctuation
from urllib.parse import unquote
from .enums import AppType
from .lazy_regex import RegexLazy, RegexLazyIgnore

PUNC_SPACE = f'{punctuation} '
trans_tbl = str.maketrans(dict.fromkeys(PUNC_SPACE, ''))
punctuation_tbl = str.maketrans(dict.fromkeys(' /.', ''))
number_table = str.maketrans(dict.fromkeys('0123456789', ''))
REPEATED_CHARACTERS = RegexLazy(r'(.)(\1{11,})')

TRIM_LONG_COMBO_EXTENSIONS = RegexLazyIgnore(
    r'^(?P<name>.{4,})\.?(Notification-?Service-?Extension|Push-?Preview|Trends-?Today-?Extension|HomeTrafficWidgetExtension|NotificationServiceAppExtension|NotiService)$'
)
TRIM_FRAMEWORK_SUFFIXES = RegexLazyIgnore(
    r'^(?P<name>.{4,})\.?((AppAndExtensions|Extension|Module|Voice)Framework)$'
)
TRIM_NOTIFICATION_EXTENSION = RegexLazyIgnore(
    r'^(?P<name>.{4,})\.?(notifications?(?:service)?-?(extension|service|content)$)'
)
TRIM_KIT_SERVICE = RegexLazyIgnore(r'(?P<name>.{4,})\.?kitservice')
TRIM_NOTIFICATION_AGENT = RegexLazyIgnore(r'(?P<name>.{4,})\.?NotificationAgent$')
TRIM_VARIOUS_EXTENSIONS = RegexLazyIgnore(
    r'^(?P<name>.{4,})\.?(about|intents?|message|share|sharing|service|today|widgets?)-?extension$'
)
TRIM_SHORT_EXTENSIONS = RegexLazyIgnore(r'^(?P<name>.{4,})\.?(extension|widgets?)$')
TRIM_FILE_PROVIDER = RegexLazyIgnore(r'^(?P<name>.{4,})\.?(fileprovider-?(extension)?)')

TRIM_SUFFIX_PATTERNS = (
    TRIM_LONG_COMBO_EXTENSIONS,
    TRIM_FRAMEWORK_SUFFIXES,
    TRIM_NOTIFICATION_EXTENSION,
    TRIM_FILE_PROVIDER,
    TRIM_KIT_SERVICE,
    TRIM_NOTIFICATION_AGENT,
    TRIM_VARIOUS_EXTENSIONS,
    TRIM_SHORT_EXTENSIONS,
)


# Safari often appends a meaningless alphanumeric string enclosed in parens.
# Otherwise, the UAs are identical so strip that suffix
# Mozilla/5.0 (iPhone; CPU iPhone OS 12_1_2 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/16C101 (5836419392)  # noqa
# Mozilla/5.0 (iPhone; CPU iPhone OS 12_1_4 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/16D57 baidumap_IPHO (10793838272)  # noqa
STRIP_NUM_SUFFIX = RegexLazyIgnore(r'(\([0-9]+\))$')
COMMON_BIGRAMS = RegexLazyIgnore(
    r'th|he|in|er|an|ar|re|on|at|en|nd|ti|es|or|te|of|ed|is|it|al|ar|st|to|nt|ng|se|'
    r'ha|as|ou|io|le|ve|co|me|de|hi|ri|ro|ic|ne|ea|ee|oo|ra|ce|li|ch|ll|be|ma|si|om|ur'
)
ILLEGAL_BIGRAMS = RegexLazyIgnore(r'[a-z0-9](jq|qg|qk|qy|qz|wq|wz)')
LEGAL_BIGRAMS = RegexLazyIgnore(
    '(aa|ba|ca|da|ea|fa|ga|ha|ia|ja|ka|la|ma|na|oa|pa|qa|ra|sa|ta|ua|va|wa|xa|ya|za|ab|bb|cb|db|eb'
    '|fb|gb|hb|ib|jb|kb|lb|mb|nb|ob|pb|qb|rb|sb|tb|ub|vb|wb|xb|yb|zb|ac|bc|cc|dc|ec|fc|gc|hc|ic|jc'
    '|kc|lc|mc|nc|oc|pc|qc|rc|sc|tc|uc|vc|wc|xc|yc|zc|ad|bd|cd|dd|ed|fd|gd|hd|id|jd|kd|ld|md|nd|od'
    '|pd|qd|rd|sd|td|ud|vd|wd|xd|yd|zd|ae|be|ce|de|ee|fe|ge|he|ie|je|ke|le|me|ne|oe|pe|qe|re|se|te'
    '|ue|ve|we|xe|ye|ze|af|bf|cf|df|ef|ff|gf|hf|if|jf|kf|lf|mf|nf|of|pf|qf|rf|sf|tf|uf|vf|wf|xf|yf'
    '|zf|ag|bg|cg|dg|eg|fg|gg|hg|ig|jg|kg|lg|mg|ng|og|pg|rg|sg|tg|ug|vg|wg|xg|yg|zg|ah|bh|ch|dh|eh'
    '|fh|gh|hh|ih|jh|kh|lh|mh|nh|oh|ph|qh|rh|sh|th|uh|vh|wh|xh|yh|zh|ai|bi|ci|di|ei|fi|gi|hi|ii|ji'
    '|ki|li|mi|ni|oi|pi|qi|ri|si|ti|ui|vi|wi|xi|yi|zi|aj|bj|cj|dj|ej|fj|gj|hj|ij|jj|kj|lj|mj|nj|oj'
    '|pj|qj|rj|sj|tj|uj|vj|wj|xj|yj|zj|ak|bk|ck|dk|ek|fk|gk|hk|ik|jk|kk|lk|mk|nk|ok|pk|rk|sk|tk|uk'
    '|vk|wk|xk|yk|zk|al|bl|cl|dl|el|fl|gl|hl|il|jl|kl|ll|ml|nl|ol|pl|ql|rl|sl|tl|ul|vl|wl|xl|yl|zl'
    '|am|bm|cm|dm|em|fm|gm|hm|im|jm|km|lm|mm|nm|om|pm|qm|rm|sm|tm|um|vm|wm|xm|ym|zm|an|bn|cn|dn|en'
    '|fn|gn|hn|in|jn|kn|ln|mn|nn|on|pn|qn|rn|sn|tn|un|vn|wn|xn|yn|zn|ao|bo|co|do|eo|fo|go|ho|io|jo'
    '|ko|lo|mo|no|oo|po|qo|ro|so|to|uo|vo|wo|xo|yo|zo|ap|bp|cp|dp|ep|fp|gp|hp|ip|jp|kp|lp|mp|np|op'
    '|pp|qp|rp|sp|tp|up|vp|wp|xp|yp|zp|aq|bq|cq|dq|eq|fq|gq|hq|iq|kq|lq|mq|nq|oq|pq|qq|rq|sq|tq|uq'
    '|vq|xq|yq|zq|ar|br|cr|dr|er|fr|gr|hr|ir|jr|kr|lr|mr|nr|or|pr|qr|rr|sr|tr|ur|vr|wr|xr|yr|zr|as'
    '|bs|cs|ds|es|fs|gs|hs|is|js|ks|ls|ms|ns|os|ps|qs|rs|ss|ts|us|vs|ws|xs|ys|zs|at|bt|ct|dt|et|ft'
    '|gt|ht|it|jt|kt|lt|mt|nt|ot|pt|qt|rt|st|tt|ut|vt|wt|xt|yt|zt|au|bu|cu|du|eu|fu|gu|hu|iu|ju|ku'
    '|lu|mu|nu|ou|pu|qu|ru|su|tu|uu|vu|wu|xu|yu|zu|av|bv|cv|dv|ev|fv|gv|hv|iv|jv|kv|lv|mv|nv|ov|pv'
    '|qv|rv|sv|tv|uv|vv|wv|xv|yv|zv|aw|bw|cw|dw|ew|fw|gw|hw|iw|jw|kw|lw|mw|nw|ow|pw|qw|rw|sw|tw|uw'
    '|vw|ww|xw|yw|zw|ax|bx|cx|dx|ex|fx|gx|hx|ix|jx|kx|lx|mx|nx|ox|px|qx|rx|sx|tx|ux|vx|wx|xx|yx|zx'
    '|ay|by|cy|dy|ey|fy|gy|hy|iy|jy|ky|ly|my|ny|oy|py|ry|sy|ty|uy|vy|wy|xy|yy|zy|az|bz|cz|dz|ez|fz'
    '|gz|hz|iz|jz|kz|lz|mz|nz|oz|pz|rz|sz|tz|uz|vz|wz|xz|yz|zz|3d|2p)'
)
COMMON_TRIGRAMS = RegexLazyIgnore(
    r'(the|and|ing|her|hat|his|tha|ere|for|lab|ent|ion|ter|was|you|ith|ver|all|wit|thi|tio|p2p|dev|'
    r'ink|jet|ios|mac|win|dos|rss|med|fun|sun|fax|app|api|bot|cam|cpu|lab|hue|rgb|sdk|web|oku|pad|'
    r'art|bea|boa|cat|law|mes|new|not|vpn|zig|4x4|gun|sim|mp3|wma|mov|pic|pix|vid|war|key|sho|syn|'
    r'bit|boy|cli|dri|han|log|man|mot|rad|oil|sip|nom|gen|per|sig|tim|tun|wor)'
)
COMMON_QUADRIGRAMS = RegexLazyIgnore(
    '(clou|ipod|ipad|that|ther|with|tion|here|ould|ight|have|hich|whic|this|thin|they|atio|ever|'
    'from|ough|were|hing|ment|droid|mobil|brows|load|micro|msdw|network|cloud|http|phone|tablet|'
    'wifi|window|page|cart|wiki|mozilla|chrome|vivaldi|view|traf|rack|game|audio|truck|block|moat|'
    'book|sport|widget|scan|connect|free|talk|school|tube|plus|tool|sheet|shop|kids|sing|date|play|'
    'client|girl|plan|proj|soft|tock|instal|linu|data|trav|stud|send|sale|phot|shar|wire|star|mail|'
    'win3|win6)'
)

UUID_LIKE_NAME = RegexLazyIgnore(
    "^([0-9a-f]{8}(-[0-9a-f]{4}){2,}$)|"
    "^([0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$)"
)
# a_05D42541-648B-41DD-B11F-0CAF61F4CE19-660-0000004D54BA56F2
# a_03E848FE-2BD9-4400-B87B-172C35D4A309-3121-00000B5202325FA0
LONG_UUID = RegexLazyIgnore(r"^([\w\d]{10}-)+([\w\d]{4,5}-)+([\w\d]{12}-)([\w\d]{3,5}-)")
SHORT_UUID = RegexLazyIgnore(r"^(\w{4,8}-)(\w{4,8}-)+(\w{4,12})$")
TWO_SEGMENT_UUID = RegexLazyIgnore(r"^(\w{4,8}-)(\w{4,8})$")
INTEGER = RegexLazyIgnore(r"\d")
MIN_WORD_LENGTH = 7


def ua_hash_key(user_agent: str, headers: dict[str, str] | None = None) -> str:
    """
    Return short hash of User Agent string for
    memory-efficient cache key.
    """
    if headers:
        try:
            cache_key = f'{user_agent}{"-".join(sorted(headers.values()))}'
        except TypeError:
            cache_key = f'{user_agent}{headers}'
    else:
        cache_key = user_agent

    return blake2s(cache_key.encode('utf-8')).hexdigest()[:12]


def long_ua_no_punctuation(user_agent: str) -> bool:
    """
    UserAgent string is long and has no Space, Dot or Slash
    """
    if len(user_agent) < 65:
        return False
    punc_removed = user_agent.translate(punctuation_tbl)
    return punc_removed == user_agent


def only_numerals_and_punctuation(user_agent: str) -> bool:
    """
    Remove all punctuation. If only digits remain,
    don't bother saving, as nothing can be learned.

    21/4.35.1.2
    5.0.6
    """
    return user_agent.translate(trans_tbl).isdigit()


def mostly_numerals(user_agent: str) -> bool:
    """
    UserAgent string is mostly numeric, discard
    15B93
    """
    if not user_agent or ' ' in user_agent:
        return False

    try:
        int(user_agent)
        return True
    except ValueError:
        pass

    alphabetic_chars = 0
    for char in user_agent:
        if not char.isnumeric():
            alphabetic_chars += 1

    return alphabetic_chars / len(user_agent) < 0.33


def clean_ua(user_agent: str, user_agent_lower: str) -> str:
    """
    Normalize and decode User Agent string
    """
    ua = unquote(STRIP_NUM_SUFFIX.sub('', user_agent)).strip()

    for prefix in (
        # sprd-Galaxy-S4/1.0 Linux/2.6.35.7 Android/4.2.2 Release/10.14.2013 Browser/AppleWebKit533.1 (KHTML, like Gecko) Mozilla/5.0 Mobile  # noqa
        # sprd-lingwin-U820S/1.0 Linux/2.6.35.7 Android/2.3.5 Release/10.15.2012 Browser/AppleWebKit533.1 (KHTML, like Gecko) Mozilla/5.0 Mobile  # noqa
        'sprd-',
        # null (FlipboardProxy/1.1; http://flipboard.com/browserproxy)
        # (null) MyOperations/3.0.0/162 JDM/1.0
        # (null)/14898/app_MUd3n2yw07Lg5hy0f8hRXuj1jI5ml17ww3haFrbKUBw/ios/2.4.0/iOS/12.1.4/gzip/s
        'null',
        '(null)',
        # AmazonWebView/Kindle for iOS/6.9.1.3/iOS/11.4.1/iPhone
        # AmazonWebView/PrimeNow/5.7/iOS/11.4.1/iPhone
        # AmazonWebView/Prime Video/5.71.1526.2/iOS/11.4.1/iPad
        # AmazonWebView/SellingServicesOnAmazon/1.1.7/iPhone OS/11.3.1/iPhone
        'amazonwebview',
        # com.usebutton.merchant/0.1.0-beta3 1 (iOS 11.4; iPhone8,4; com.tophatter.tophatter/4.33 614; Scale/2.00; en-US)
        # com.usebutton.merchant/1.0.0 1 (Android 6.0.1; samsung SM-G900V; com.walmart.grocery/7.1.0 7010103; Scale/3.0; en_us)
        # com.usebutton.merchant/1.0.0 1 (iOS 13.2.3; iPhone12,1; com.giddyinc.boxed.prod.00/2.9.4 2.9.4.11504; Scale/2.00; en-US)
        # com.usebutton.merchant/1.0.0 1 (iOS 13.2.3; iPhone12,1; com.houzz.app/19.12.15 4305; Scale/2.00; en-US)
        # com.usebutton.mparticle/7.9.2-1.0.0 (iOS 13.3; iPhone8,1; com.walmart.grocery/2001081028; Scale/2.0; en-US)
        # com.usebutton.mparticle/7.9.2-1.0.0 (iOS 13.3; iPhone9,1; com.walmart.electronics/2003032106; Scale/2.0; en-US)
        # com.usebutton.sdk/5.32.0 (iOS 12.2; iPhone10,1; com.hotwire.Hotwire/1732244; Scale/2.00; en-US; Build/appstore)
        # com.usebutton.sdk/5.32.0 (iOS 12.2; iPhone10,3; doordash.DoorDashConsumer/1; Scale/3.00; en-US; Build/appstore)
        'com.usebutton.',
        # FirebaseAuth.iOS/10.14.0 com.samsclub.samsapp/25.10.40 iPhone/26.0.1 hw/iPhone18_3 iOS 10.14.0 - Sam's Club - iPhone 17
        # FirebaseAuth.iOS/7.7.0 com.storytree.businessprints/4.7.9 iPhone/17.5.1 hw/iPhone12_1
        # FirebaseAuth.iOS/8.4.0 depollsoft.pitchperfect/2.0.3 iPhone/18.4 hw/iPhone14_6
        # FirebaseAuth.iOS/9.3.0 com.2mc.bidftamobile/4.26 iPhone/18.6 hw/iPhone12_8
        'firebaseauth.ios',
    ):
        if user_agent_lower.startswith(prefix):
            ua = ua[len(prefix) :].strip().strip('/')

    return ua


def mostly_repeating_characters(user_agent: str) -> bool:
    """
    User Agent string is mostly repeating characters
      - baaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa
      - 00000000-0000-0000-0000-000000000000
    """
    if not user_agent or len(user_agent) > 100:
        return False

    counts = Counter(user_agent)
    return len(counts) / len(user_agent) < 0.15


def random_alphanumeric_string(user_agent: str) -> bool:
    """
    User Agent string is mostly generated names like:

    001471FmBjtgZvahkMJdcGJhhjXuuD99
    002261f23pRwrjIEtS1MrX0lZ4hx8N7P
    002353ueFfucaEDjKRbKwlpDpecxwYwC
    ziNICEarE9VlaPSkhDAyZrkZSpuEkIA
    vVNYZaiXO9Hd5zAi
    """
    if not user_agent or not user_agent.isascii() or good_ngram_matches(user_agent):
        return False

    # holav1_10593F39DD2DAEBC
    # HKUM_XBw3S
    # HKUM_fXHH8t9R
    if user_agent.startswith(('holav1_', 'hkum_')):
        return True

    # strings with adequate spaces / punctuation are not handled
    if (ua_length := len(user_agent)) <= MIN_WORD_LENGTH:
        return False

    if well_punctuated(user_agent, ua_length):
        return False

    vowels = 'aeiou'
    previous_is_int = user_agent[0].isnumeric()

    intermingled_integers = 1 if previous_is_int else 0

    longest_consonant_sequence = 0
    longest_vowel_sequence = 0

    current_consonant_sequence = 0
    current_vowel_sequence = 0

    for char in user_agent:
        char_is_numerical = char.isnumeric()

        if char_is_numerical:
            if not previous_is_int:
                intermingled_integers += 1
            reset_vowels = reset_consonants = previous_is_int = True

        elif char in vowels:
            current_vowel_sequence += 1
            reset_vowels, reset_consonants = False, True

        else:
            current_consonant_sequence += 1
            reset_vowels, reset_consonants = True, False

        if reset_vowels:
            current_vowel_sequence, longest_vowel_sequence = reset_sequences(
                current_vowel_sequence,
                longest_vowel_sequence,
            )

        if reset_consonants:
            current_consonant_sequence, longest_consonant_sequence = reset_sequences(
                current_consonant_sequence,
                longest_consonant_sequence,
            )

        if not char_is_numerical and previous_is_int:
            previous_is_int = False

    # update longest sequences after all chars are seen
    current_consonant_sequence, longest_consonant_sequence = reset_sequences(
        current_consonant_sequence,
        longest_consonant_sequence,
    )
    current_vowel_sequence, longest_vowel_sequence = reset_sequences(
        current_vowel_sequence,
        longest_vowel_sequence,
    )

    short_ua_len = 17
    alphabetic_threshold = 4 if ua_length < short_ua_len else 5
    intermingled_integer_threshold = 2 if ua_length < short_ua_len else 3

    gibberish = (
        longest_consonant_sequence >= alphabetic_threshold
        or longest_vowel_sequence >= alphabetic_threshold
        or intermingled_integers >= intermingled_integer_threshold
    )

    if gibberish and ua_length < short_ua_len:
        return ngram_analysis_gibberish(user_agent)

    return gibberish or ngram_analysis_gibberish(user_agent)


def well_punctuated(user_agent: str, ua_length: int) -> bool:
    """
    User Agents should have a number of spaces & punctuation characters.
    UAs with adequate spaces / punctuation are not gibberish.
    """
    punctuation_removed_diff = ua_length - len(user_agent.translate(trans_tbl))
    if ua_length <= 20:
        delta = 1
    elif ua_length <= 35:
        delta = 2
    else:
        delta = 3

    return punctuation_removed_diff >= delta


def ngram_analysis_gibberish(user_agent: str) -> bool:
    """
    Analyze legal ngrams to see if it's likely that UA is gibberish.
    """
    ua_length = len(user_agent)

    if ua_length <= 6:
        bigram_threshold = 3
    elif ua_length <= 10:
        bigram_threshold = 4
    elif ua_length <= 15:
        bigram_threshold = 5
    elif ua_length <= 20:
        bigram_threshold = 7
    elif ua_length <= 30:
        bigram_threshold = 10
    else:
        bigram_threshold = 15

    legal_bigram_matches = LEGAL_BIGRAMS.findall(user_agent)
    unique_bigrams = len(legal_bigram_matches)

    return bigram_threshold > unique_bigrams


def good_ngram_matches(user_agent: str) -> bool:
    """
    User agent likely NOT gibberish due to containing
    quadrigrams or trigrams.

    Quadrigrams:
       - Wordswithfriends3
       - ZooskAndroid

    Trigrams:
       - ZipsCarWash
       - JQSmartBand
    """
    if COMMON_QUADRIGRAMS.search(user_agent):
        return True

    trigram_match = COMMON_TRIGRAMS.findall(user_agent)
    ua_length = len(user_agent)

    if ua_length < 24 and trigram_match or len(trigram_match) >= 2:
        return True

    if ua_length <= 12:
        bigram_threshold = 2
    elif ua_length <= 24:
        bigram_threshold = 4
    elif ua_length <= 36:
        bigram_threshold = 6
    else:
        bigram_threshold = 8

    common_bigram_match = COMMON_BIGRAMS.findall(user_agent)

    return len(common_bigram_match) >= bigram_threshold


def reset_sequences(current: int, longest: int) -> tuple[int, int]:
    """
    Set longest sequence to current if it's shorter
    and reset current sequence to zero.
    """
    return 0, current if current > longest else longest


def uuid_like_name(value: str) -> bool:
    """
    Ensure that name isn't UUID-like string, such as:

    738FAAEF-30CF
    6BFAD903-A5EA-4E34
    5FAEB6ED-AE46-4A26-BA1B
    ea1866cb-c89a-6d5d-89b8-afdcdb715237
    """
    if not INTEGER.search(value):
        return False

    if UUID_LIKE_NAME.search(value) or LONG_UUID.search(value) or SHORT_UUID.search(value):
        return True

    # all sections of the string must contain an integer
    try:
        for group in TWO_SEGMENT_UUID.search(value).groups():
            if not INTEGER.search(group):
                return False
        return True
    except Exception:
        pass

    return False


def calculate_dtype(app_name: str) -> AppType:
    """
    For generic extractors try to return a more
    specific type we can be if reasonably sure.
    """
    app_name_lower = app_name.lower()
    for name, dtype in (
        ('update', AppType.DesktopApp),
        ('mail', AppType.PIM),
        ('api', AppType.Library),
        ('sdk', AppType.Library),
        ('webview', AppType.Browser),
    ):
        if name in app_name_lower:
            return dtype

    return AppType.Generic


def normalize_app_name(app_name: str) -> str:
    """
    Remove suffixes such as:

    com.costco.costco.notification-extension
    BarcelonaNotificationExtension
    OneDriveNotificationServiceExtension
    TWorldNotificationService
    SpyPoint.notificationsContent
    RampNotificationServiceExtension

    com.google.Drive.ModuleFramework
    com.google.Drive.ExtensionFramework
    """
    # endswith check is almost free when a regex will match,
    # and 100x faster when no match will be found.
    if not app_name.lower().endswith((
        'agent',
        'content',
        'extension',
        'extensions',
        'framework',
        'preview',
        'provider',
        'service',
        'widget',
        'widgets',
    )):
        return app_name

    for pattern in TRIM_SUFFIX_PATTERNS:
        if matched := pattern.match(app_name):
            return matched.group('name').strip('.')
    return app_name


__all__ = (
    'calculate_dtype',
    'ua_hash_key',
    'long_ua_no_punctuation',
    'only_numerals_and_punctuation',
    'mostly_numerals',
    'clean_ua',
    'mostly_repeating_characters',
    'random_alphanumeric_string',
    'uuid_like_name',
    'normalize_app_name',
)



from .base import BaseDeviceParser
from ua_extract.enums import DeviceType
from typing import Any
from ...lazy_regex import RegexLazyIgnore
from ...settings import BOUNDED_REGEX, DDCache

HBBTV_FRAGMENT = RegexLazyIgnore(r'(?:HbbTV|SmartTvA)/([1-9]{1}(?:\.[0-9]{1}){1,2})')
SHELL_TV_FRAGMENT = RegexLazyIgnore(r'[ _]Shell[ _]\w{6}|tclwebkit(\d+[.\d]*)')


class BaseTvParser(BaseDeviceParser):
    DEVICE_TYPE = DeviceType.TV
    __slots__ = (
        '_is_hbbtv',
        '_is_shell_tv',
    )

    def __init__(self, *args, **kwargs):  # type: ignore[no-untyped-def]
        super().__init__(*args, **kwargs)

        self._is_hbbtv: bool | None = None
        self._is_shell_tv: bool | None = None

    def is_hbbtv(self) -> bool:
        """
        HbbTV UA strings are only parsed by the televisions.yml file
        """
        if not self._is_hbbtv:
            self._is_hbbtv = HBBTV_FRAGMENT.search(self.user_agent) is not None
        return self._is_hbbtv

    def is_shell_tv(self) -> bool:
        """
        Shell UA strings are only parsed by the shell_tv.yml file
        """
        if not self._is_shell_tv:
            self._is_shell_tv = SHELL_TV_FRAGMENT.search(self.user_agent) is not None
        return self._is_shell_tv

    def set_device_type(self) -> None:
        """
        Set device type, at least,since we know this is a TV.
        """
        if not self.ua_data:
            self.ua_data = {
                'brand': '',
                'model': '',
                'type': self.dtype(),
            }


class HbbTv(BaseTvParser):
    __slots__ = ()

    @property
    def regex_list(self) -> list[dict[str, Any]]:
        cache_key = 'tv_regexes'
        if regexes := DDCache.get(cache_key, []):
            return regexes

        regexes = self.load_from_yaml('regexes/upstream/device/televisions.yml')
        if not regexes:
            return []

        reg_list = []
        for brand, stats in regexes.items():
            brand_data = {
                'brand': brand,
                'regex': RegexLazyIgnore(BOUNDED_REGEX.format(stats['regex'])),
                'device': stats['device'],
            }
            if 'models' in stats:
                for model in stats['models']:
                    model['regex'] = RegexLazyIgnore(BOUNDED_REGEX.format(model['regex']))
                brand_data['models'] = stats['models']
            if 'model' in stats:
                brand_data['model'] = stats['model']
            reg_list.append(brand_data)

        DDCache[cache_key] = reg_list

        return reg_list

    def _parse(self) -> None:
        if self.is_hbbtv():
            super()._parse()
            return self.set_device_type()


class ShellTv(BaseTvParser):
    __slots__ = ()
    fixture_files = [
        'upstream/device/shell_tv.yml',
    ]

    def _parse(self) -> None:
        if self.is_shell_tv():
            super()._parse()
            return self.set_device_type()


__all__ = (
    'HbbTv',
    'ShellTv',
)



from typing import Any
from . import BaseDeviceParser
from ..lazy_regex import RegexLazyIgnore


class OSFragment(BaseDeviceParser):
    fixture_files = [
        'local/osfragments.yml',
    ]

    def yaml_to_list(self, yfile: str) -> list[dict[str, Any]]:
        """
        List of dicts like so:

        {'name': 'iOS', 'regexes': ['iPhone', 'iPad']}
        """
        new_regexes = self.load_from_yaml(yfile)
        reg_list = []

        # load and compile regexes. Not using boundaries.
        for os, regexes in new_regexes.items():  # type: ignore[union-attr]
            reg_list.append({
                'name': os,
                'regexes': [RegexLazyIgnore(reg) for reg in regexes],
            })

        return reg_list

    def _parse(self) -> None:
        for ua_data in self.regex_list:
            for regex in ua_data['regexes']:
                matched = regex.search(self.user_agent)

                if matched:
                    self.matched_regex = matched
                    self.ua_data['name'] = ua_data['name']
                    self.known = True

                    return


__all__ = [
    'OSFragment',
]



from urllib.parse import unquote
import regex
from regex import IGNORECASE


# When one of these attributes is called, compile the regex
REGEX_ATTRS = {
    'match',
    'fullmatch',
    'search',
    'sub',
    'subf',
    'subfn',
    'split',
    'splititer',
    'findall',
    'finditer',
    'purge',
    'escape',
    'compiled',
}


class RegexLazy:
    """
    Defer compilation of regex until it's actually called.
    Some regexes, especially on device models will almost
    never be called, so save the compilation time.
    """

    def __init__(self, pattern: str, flags: int = 0) -> None:
        # Decode UA regexes because UA strings are also decoded
        # Pic%20Collage/(\d+[\.\d]+) CFNetwork
        self.pattern = unquote(pattern)
        self.flags = flags
        self.compiled = None

    def __getattribute__(self, attribute: str) -> regex.Regex:
        compiled_regex = super().__getattribute__('compiled')
        if compiled_regex is None and attribute in REGEX_ATTRS:
            pattern = super().__getattribute__('pattern')
            flags = super().__getattribute__('flags')
            compiled_regex = regex.compile(pattern, flags)
            self.compiled = compiled_regex

        if attribute == 'compiled':
            return compiled_regex

        return getattr(compiled_regex, attribute)

    def __repr__(self) -> str:
        return repr(self.compiled)

    def __hash__(self) -> int:
        return hash(self.compiled)

    def __eq__(self, other: regex.Regex) -> bool:
        return self.compiled == other.compiled


class RegexLazyIgnore(RegexLazy):
    def __init__(self, pattern: str) -> None:
        super().__init__(pattern, IGNORECASE)


__all__ = (
    'RegexLazy',
    'RegexLazyIgnore',
)



from .base import *
from .adobe_cc import *
from .browser import *
from .desktopapp import *
from .dictua import *
from .feed_reader import *
from .extractor_name_version import *
from .extractor_whole_name import *
from .library import *
from .mediaplayer import *
from .messaging import *
from .mobileapp import *
from .pim import *
from .vpnproxy import *



from ..base import ParserBaseTest
from ...device_detector import DeviceDetector


class TestCache(ParserBaseTest):
    """
    Values should remain the same on multiple runs when
    the parsing on followup runs pulls from cached values.
    """

    def test_is_bot(self):
        ua = "Mozilla/5.0 (compatible; CloudFlare-AlwaysOnline/1.0; +http://www.cloudflare.com/always-online) AppleWebKit/534.34"

        first_run = DeviceDetector(ua).parse()
        self.assertTrue(first_run.is_bot())

        second_run = DeviceDetector(ua).parse()
        self.assertTrue(second_run.is_bot())

    def test_device(self):
        ua = "Mozilla/5.0 (Linux; U; Android 2.3.3; ja-jp; COOLPIX S800c Build/CP01_WW) AppleWebKit/533.1 (KHTML, like Gecko) Version/4.0 Mobile Safari/533.1"

        first_run = DeviceDetector(ua).parse()
        self.assertEqual(first_run.device_type(), 'camera')
        self.assertEqual(first_run.device_brand(), 'Nikon')

        second_run = DeviceDetector(ua).parse()
        self.assertEqual(second_run.device_type(), 'camera')
        self.assertEqual(second_run.device_brand(), 'Nikon')

    def test_client(self):
        ua = "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/80.0.3987.162 Safari/537.36"

        first_run = DeviceDetector(ua).parse()
        self.assertEqual(first_run.client_name(), 'Chrome')
        self.assertEqual(first_run.client_version(), '80.0.3987.162')

        second_run = DeviceDetector(ua).parse()
        self.assertEqual(second_run.client_name(), 'Chrome')
        self.assertEqual(second_run.client_version(), '80.0.3987.162')

    def test_os(self):
        ua = "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:74.0) Gecko/20100101 Firefox/74.0"

        first_run = DeviceDetector(ua).parse()
        self.assertEqual(first_run.os_name(), 'Ubuntu')

        second_run = DeviceDetector(ua).parse()
        self.assertEqual(second_run.os_name(), 'Ubuntu')


__all__ = [
    'TestCache',
]



from urllib.parse import unquote
from ua_extract.parser import ClientHints
from ..base import GenericParserTest, ParserBaseTest
from ...parser import (
    AdobeCC,
    Browser,
    DictUA,
    FeedReader,
    Library,
    MediaPlayer,
    MobileApp,
    PIM,
    NameVersionExtractor,
    WholeNameExtractor,
)


class ParserClientBase(ParserBaseTest):

    def test_parsing(self):
        if not self.Parser or not self.fields:
            return

        fixtures = self.load_fixtures()

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            expect = fixture[self.fixture_key]
            parsed = self.Parser(
                self.user_agent,
                client_hints=ClientHints.new(fixture.get('headers', {})),
            ).clear_cache().parse()  # clear cache because fixture files may contain duplicate UAs
            data = parsed.ua_data

            for field in self.fields:
                expected = str(expect.get(field, '')).lower()
                extracted = str(data.get(field, '')).lower()

                # Consider names valid if they vary only by casing and spacing
                if field == 'name' and extracted:
                    if (
                        expected != extracted
                        and expected.replace(' ', '') == extracted.replace(' ', '')
                    ):
                        continue

                self.assertIn(
                    field,
                    data,
                    msg=f'Error parsing {self.user_agent!r}. Parsed data missing {field!r} key.'
                )
                self.assertEqual(expected, extracted, field=field)


class TestBrowser(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/client/browser.yml',
        'tests/parser/fixtures/local/client/browser.yml',
    ]
    fields = ('name', 'type')
    Parser = Browser


class TestAdobeCC(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/local/client/adobe_cc.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = AdobeCC


class TestDictUA(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/local/client/dictua.yml',
    ]
    fields = ('name', 'version')
    Parser = DictUA


class TestFeedReader(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/client/feed_reader.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = FeedReader


class TestLibrary(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/local/client/library.yml',
        'tests/parser/fixtures/upstream/client/library.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = Library


class TestMediaPlayer(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/client/mediaplayer.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = MediaPlayer


class TestMobileApp(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/client/mobile_app.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = MobileApp


class TestPIM(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/upstream/client/pim.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = PIM


class TestNameVersionExtractor(ParserClientBase):

    fixture_files = [
        'tests/parser/fixtures/local/client/extractor_name_version.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = NameVersionExtractor
    skipped = (
        '$(PRODUCT_NAME)/4839 CFNetwork/894 Darwin/17.4.0',
        'MotionXGPSFull24.2b5063R-iOS12.0-iPhone7,2',
        '1530819907iOSv1.6.5',
        'ISUA11.00MP',
    )


class TestWholeNameExtractor(GenericParserTest):

    fixture_files = [
        'tests/parser/fixtures/local/client/extractor_whole_name.yml',
    ]
    fields = ('name', 'type', 'version')
    Parser = WholeNameExtractor
    skipped = [
        '646E514C51BFF23DBBB5B9487F142670_5_5.10.228257.2253_1_0',
    ]

    def test_ua_with_no_interesting_keys(self):
        user_agents = (
            'Chrome/136.0.7103.42 iPhone/18.6.2 hw/iPhone14_5',
            'Chrome/138.0.7204.156 iPad/18.6.2 hw/iPad14_',
        )
        for ua in user_agents:
            wn = WholeNameExtractor(ua, client_hints=None).parse()
            self.assertEqual(wn.name(), '')


# class TestNoNameExtracted(ParserBaseTest):
#
#     fixture_files = [
#         'tests/parser/fixtures/local/client/extractor_no_name.yml',
#     ]
#
#     def test_parsing(self):
#         fixtures = self.load_fixtures()
#         fields = ('name', 'version')
#         extractors = (WholeNameExtractor, NameVersionExtractor)
#
#         for fixture in fixtures:
#             ua = fixture.pop('user_agent')
#
#             for extractor in extractors:
#                 parsed = extractor(ua, ua_hash(ua)).parse()
#
#                 for field in fields:
#                     self.assertNotIn(field, parsed.ua_data)

__all__ = (
    'TestBrowser',
    'TestFeedReader',
    'TestLibrary',
    'TestMediaPlayer',
    'TestMobileApp',
    'TestPIM',
    'TestNameVersionExtractor',
    'TestWholeNameExtractor',
    # 'TestNoNameExtracted',
)



from urllib.parse import unquote
from ..base import ParserBaseTest
from ...parser import OS, OSFragment, ClientHints


class TestOS(ParserBaseTest):

    fixture_files = [
        'tests/parser/fixtures/local/oss.yml',
        'tests/parser/fixtures/upstream/oss.yml',
    ]
    fields = ('name', 'version')
    fixture_key = 'os'
    Parser = OS


class TestOSFragment(ParserBaseTest):

    fixture_files = [
        'tests/parser/fixtures/local/osfragments.yml',
    ]
    fields = ('name',)
    fixture_key = 'name'

    def test_parsing(self):
        fixtures = self.load_fixtures()

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            expect = fixture['name']
            ch = ClientHints.new(fixture.get('headers', {}))
            parsed = OSFragment(self.user_agent, client_hints=ch).parse()

            self.assertEqual(expect, parsed.ua_data['name'])


__all__ = (
    'TestOS',
    'TestOSFragment',
)



from . import BaseClientParser
from ua_extract.enums import AppType
from ...lazy_regex import RegexLazyIgnore

ADOBE_CC = RegexLazyIgnore(r'\b(?P<name>[\w ]+)/(?P<version>[\.\d]+) com\.adobe\.[\w+\.-]+/')


class AdobeCC(BaseClientParser):
    """
    Check Adobe Creative Cloud apps.
    """

    CHECK_APP_ID = False
    __slots__ = ()
    APP_TYPE = AppType.DesktopApp

    def _parse(self) -> None:
        if adobe_app := ADOBE_CC.search(self.user_agent):
            name, version = adobe_app.groups()
            if not name.lower().startswith('adobe'):
                name = f'Adobe {name}'
            self.app_name = name
            self.app_version = version
            self.ua_data = {
                'name': name,
                'version': version,
                'type': self.APP_TYPE,
            }


__all__ = [
    'AdobeCC',
]



from ..lazy_regex import RegexLazyIgnore
from ..yaml_loader import RegexLoader, app_pretty_names_types_data
from ua_extract.enums import AppType
from ua_extract.utils import normalize_app_name

APP_ID_SANS_VERSION = RegexLazyIgnore(r'\b([a-z]{1,5}(?:\.[\w-]+){1,})')

# 6H4HRTU5E3.com.avast.osx.secureline.avastsecurelinehelper/47978 CFNetwork/976 Darwin/18.2.0 (x86_64)
# YMobile/1.0(com.kitkatandroid.keyboard/4.3.2;Android/6.0.1;lv1;LGE;LG-M153;;792x480
# x86_64; macOS 10.14.5 (18F132); com.apple.ap.adprivacyd; 143441-1,13
APP_ID_VERSION = RegexLazyIgnore(
    r'\b(?P<name>[a-z]{2,5}\.[\w\.\-]+)[;:/] ?(?P<version>[\d\.\-]+)\b'
)

# Match Application IDs like:
# FirebaseAuth.iOS/7.6.0 YanFlex.CPlus.Craigslist/5.1.6 iPhone/18.6 hw/iPhone15_3
# FirebaseAuth.iOS/8.4.0 depollsoft.pitchperfect/2.0.3 iPhone/18.4 hw/iPhone14_6
# but do not match domain names
LONG_PREFIX_APP_ID_VERSION = RegexLazyIgnore(
    r'[\s\(;:](?P<name>[a-z]{6,}\.[\w\.\-]{6,})[;:/] ?(?P<version>[\d\.\-]+)\b'
)


class ApplicationIDExtractor(RegexLoader):
    """
    Extract App Store IDs such as:

    extract APP IDs such as
    com.cloudveil.CloudVeilMessenger
    com.houzz.app
    com.google.Maps
    """

    key = 'app_id'

    def __init__(self, user_agent: str) -> None:
        self.user_agent = user_agent
        self.details: dict[str, str] = {}
        self._app_id_pretty_names = app_pretty_names_types_data()

    def extract(self) -> 'ApplicationIDExtractor':
        """
        Parse for

        "<tld>.<string.<string>.<string><sep><version>" or
        "<tld>.<string.<string>.<string>"

        In the (unlikely) event that multiple valid IDs
        are found, just return the first one.
        """
        if self.details:
            return self

        if not (app_ids := self.match_regexes()):
            return self

        pretty_names = self._app_id_pretty_names
        for app_id, version in app_ids:
            normalized_app_id = normalize_app_name(app_id)
            if pretty_name := pretty_names.get(normalized_app_id.lower()):
                self.details = {
                    'name': pretty_name['name'],
                    'app_id': normalized_app_id,
                    'version': version,
                    'type': pretty_name['type'],
                }
                break
        else:
            self.details = {
                'app_id': normalize_app_name(app_ids[0][0]),
                'version': app_ids[0][1],
                'type': AppType.Generic,
            }

        return self

    def match_regexes(self) -> list[tuple[str, str]]:
        """
        Extract App IDs from the user agent.
        """
        # Skip UAs like sentry.java.android.react-native
        if self.user_agent.startswith('sentry.'):
            return []

        if app_ids := LONG_PREFIX_APP_ID_VERSION.findall(self.user_agent):
            return app_ids

        if app_ids := APP_ID_VERSION.findall(self.user_agent):
            return app_ids

        if app_ids := [(app_id, '') for app_id in APP_ID_SANS_VERSION.findall(self.user_agent)]:
            return app_ids

        return []

    def version(self) -> str:
        return self.details.get('version', '')

    def pretty_name(self) -> str:
        return self.details.get('name', '')

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self.user_agent!r})'

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self.user_agent!r})'


__all__ = [
    'ApplicationIDExtractor',
]
